import shutil
from enum import Enum
from pathlib import Path
from typing import Tuple, List

from fauxpy.fault_localization.collect_mbfl.api_lib import CollectMbflApi
from fauxpy.fault_localization.granularity.db_manager import FunctionLevelDbManager
from fauxpy.fault_localization.granularity.function_level import (
    FunctionLevelGranularity,
)
from fauxpy.fault_localization.mbfl.mutation_lib.mutant import Mutant
from fauxpy.fault_localization.mbfl.db_manager import MbflDbManager
from fauxpy.fault_localization.util.path_util import PathUtil
from fauxpy.fault_localization.util.temp_lib import TempManager
from fauxpy.session_lib import naming_lib
from fauxpy.session_lib.fauxpy_path import FauxpyPath
from fauxpy.session_lib.fauxpy_printer import fl_print
from fauxpy.session_lib.fl_type import FlGranularity


class TestComparisonState(Enum):
    Normal = 0
    BadTest = 1
    BadMutant = 2
    NotTarget = 3


class MbflRunManager:
    def __init__(
        self,
        db_manager: MbflDbManager,
        function_level_db_manager: FunctionLevelDbManager,
        project_working_directory: Path,
    ):
        self._db_manager = db_manager
        self._collect_mbfl_api = CollectMbflApi()
        self._function_level_granularity_manager = FunctionLevelGranularity(
            function_level_db_manager
        )
        self._temp_manager = TempManager(project_working_directory)
        self._path_util = PathUtil(project_working_directory)

    @staticmethod
    def _backup_file(file_path: Path):
        new_path = f"{str(file_path.resolve())}.bak"
        file_path.rename(new_path)

    @staticmethod
    def _un_backup_file(file_path: Path):
        assert file_path.exists()
        backup_path_str = f"{str(file_path.resolve())}.bak"
        backup_path = Path(backup_path_str)
        assert backup_path.exists()
        file_path.unlink()
        assert not file_path.exists()
        file_path_str = str(file_path.resolve())
        backup_path.rename(file_path_str)
        assert file_path.exists()
        assert not backup_path.exists()

    @staticmethod
    def _insert_mutation_module(module_path: str, module_content: str):
        with open(module_path, "w") as file:
            file.write(module_content)

    def _mutate_temp_project(
        self,
        temp_project_path: str,
        mutating_module_path: str,
        mutating_module_content: str,
    ):
        relative_path_of_mutating_module = (
            self._path_util.absolute_path_to_relative_path(mutating_module_path)
        )
        absolute_temp_path_of_mutating_module = (
            Path(temp_project_path) / relative_path_of_mutating_module
        )
        self._backup_file(absolute_temp_path_of_mutating_module)
        self._insert_mutation_module(
            str(absolute_temp_path_of_mutating_module.resolve()),
            mutating_module_content,
        )

    def _un_mutate_temp_project(
        self, temp_project_path: str, mutating_module_path: str
    ):
        relative_path_of_mutating_module = (
            self._path_util.absolute_path_to_relative_path(mutating_module_path)
        )
        absolute_temp_path_of_mutating_module = (
            Path(temp_project_path) / relative_path_of_mutating_module
        )
        self._un_backup_file(absolute_temp_path_of_mutating_module)

    # TODO: probably, the timeout computation is not need. If timeout happens,
    #  the test cases jason file does not get generated, which is
    #  something that needs to be checked to see if timeout happened or not.
    @staticmethod
    def _compare_two_test_case_results(
        normal_test_case, mutant_test_case
    ) -> Tuple[int, int, int, TestComparisonState]:
        failed_to_passed = 0
        passed_to_failed = 0
        failed_changed = 0

        mutant_test_case_timeout_stat = mutant_test_case[3]

        normal_test_type = normal_test_case[1]
        normal_test_timeout_stat = normal_test_case[3]
        normal_test_target_stat = normal_test_case[4]
        normal_test_included = (normal_test_type == "passed") or (
            normal_test_type == "failed" and normal_test_target_stat == 1
        )

        # ToDo: use the condition inside the if. Not as a boolean variable.
        if not normal_test_included:
            # The failing test is not a target failing test. Discard the current test.
            test_comparison_state = TestComparisonState.NotTarget
        else:
            if normal_test_timeout_stat == -1 and mutant_test_case_timeout_stat == -1:
                # It's OK! Compute the terms.
                test_comparison_state = TestComparisonState.Normal
            elif normal_test_timeout_stat == -1 and mutant_test_case_timeout_stat == 1:
                # Infinite loop. Do not compute the terms. Discard current mutant.
                test_comparison_state = TestComparisonState.BadMutant
            elif normal_test_timeout_stat == 1 and mutant_test_case_timeout_stat == -1:
                # Weired case. Do not compute the terms. Discard current test.
                test_comparison_state = TestComparisonState.BadTest
            elif normal_test_timeout_stat == 1 and mutant_test_case_timeout_stat == 1:
                # Bad initial test. Do not compute the terms. Discard current test.
                test_comparison_state = TestComparisonState.BadTest
            else:
                raise Exception("This one must never happen.")

        if test_comparison_state == TestComparisonState.Normal:
            if normal_test_case[1] == "passed" and mutant_test_case[1] == "failed":
                passed_to_failed = 1
            elif normal_test_case[1] == "failed" and mutant_test_case[1] == "passed":
                failed_to_passed = 1
            elif normal_test_case[1] == "failed" and mutant_test_case[1] == "failed":
                if normal_test_case[2] != mutant_test_case[2]:
                    failed_changed = 1

        return failed_to_passed, passed_to_failed, failed_changed, test_comparison_state

    @staticmethod
    def _remove_temp_project(temp_project_path):
        shutil.rmtree(temp_project_path)

    def _get_mutant_score_terms(self, mutant_test_case_run_result_table):
        """
        Bad tests are removed from computation.
        In case of timeout_happened = True, the terms failed_to_passed, passed_to_failed, failed_changed are not valid.
        """
        failed_to_passed = 0
        passed_to_failed = 0
        failed_changed = 0
        timeout_happened = False

        for mutantTestCaseResult in mutant_test_case_run_result_table:
            normal_test_case_result = self._db_manager.select_test_case(
                mutantTestCaseResult[0]
            )

            if normal_test_case_result is None:
                # In pandas 47, one of the parametrized tests can have names
                # generated randomly (a very rare case). Thus, the collect mode
                # can return test names that do not exist in the main execution of the
                # MBFL method. In these cases, let's just ignore that test
                # and continue the score terms computation using the remaining tests.
                continue

            f2p, p2f, fc, test_comp_state = self._compare_two_test_case_results(
                normal_test_case_result, mutantTestCaseResult
            )
            if test_comp_state == TestComparisonState.Normal:
                failed_to_passed += f2p
                passed_to_failed += p2f
                failed_changed += fc
            elif (
                test_comp_state == TestComparisonState.BadTest
                or TestComparisonState.NotTarget
            ):
                continue
            elif test_comp_state == TestComparisonState.BadMutant:
                timeout_happened = True
                break

        return failed_to_passed, passed_to_failed, failed_changed, timeout_happened

    def run_all_mutants_store_db(
        self,
        mutants: List[Mutant],
        file_or_dir: List[str],
        fl_granularity: FlGranularity,
        target_src: FauxpyPath,
        exclude_list: List[FauxpyPath],
        timeout_limit: float,
        number_all_tests,
        process_timeout,
    ):
        temp_project_path = self._temp_manager.make_project_copy_in_temp()

        number_of_all_mutants = len(mutants)

        fl_print.normal(f"Running {number_of_all_mutants} Mutants")
        for index, mutant in enumerate(mutants):
            fl_print.normal(
                f"Running Mutant {mutant.get_id()} ({index + 1}/{number_of_all_mutants})"
            )

            self._mutate_temp_project(
                temp_project_path, mutant.get_module_path(), mutant.get_module_content()
            )
            mutant_test_case_run_result_table = (
                self._collect_mbfl_api.run_mbfl_collect_mode(
                    target_src,
                    exclude_list,
                    temp_project_path,
                    file_or_dir,
                    timeout=timeout_limit,
                    process_timeout=process_timeout,
                )
            )

            if (
                mutant_test_case_run_result_table is None
                or len(mutant_test_case_run_result_table) == 0
            ):
                # Timeout happened
                fl_print.normal("Timeout or bad mutant")
                self._db_manager.update_mutant_as_timeout(mutant.get_id())
                self._un_mutate_temp_project(
                    temp_project_path, mutant.get_module_path()
                )
                continue

            # For some reason, some mutant executions might return fewer
            # number of test cases. For now I remove them.
            # TODO: Find the reason. Found in sonic3. Takes around three hours.
            if number_all_tests != len(mutant_test_case_run_result_table):
                fl_print.normal("Missing tests mutant")
                self._db_manager.update_mutant_as_having_missing_tests(mutant.get_id())
                self._un_mutate_temp_project(
                    temp_project_path, mutant.get_module_path()
                )
                continue

            (
                failed_to_passed,
                passed_to_failed,
                failed_changed,
                timeout_happened,
            ) = self._get_mutant_score_terms(mutant_test_case_run_result_table)

            # In case of timeout, the mutant is thrown away and
            # its status is updated in the database
            # TODO: most probably this part is not needed
            if timeout_happened:
                self._db_manager.update_mutant_as_timeout(mutant.get_id())
            else:
                if fl_granularity == FlGranularity.Statement:
                    entity_name = naming_lib.get_statement_name(
                        mutant.get_module_path(), mutant.get_line_number()
                    )
                elif fl_granularity == FlGranularity.Function:
                    covered_function = (
                        self._function_level_granularity_manager.get_covered_function(
                            mutant.get_module_path(), mutant.get_line_number()
                        )
                    )
                    entity_name = naming_lib.get_covered_function_name(
                        covered_function[0],
                        covered_function[1],
                        covered_function[2],
                        covered_function[3],
                    )
                else:
                    # TODO: Safe to remove (handled by input validation.)
                    raise Exception(f"Granularity {fl_granularity.name} is not supported.")

                self._db_manager.insert_mutant_score_terms(
                    mutant.get_id(),
                    entity_name,
                    failed_to_passed,
                    passed_to_failed,
                    failed_changed,
                )

            self._un_mutate_temp_project(temp_project_path, mutant.get_module_path())
        self._remove_temp_project(temp_project_path)



from typing import List

from fauxpy.session_lib import naming_lib


class TargetFailingTests(object):
    def __init__(self, target_failing_tests_list: List[str]):
        self._targeted_failing_test_list = target_failing_tests_list

    def get_failing_tests(self):
        return self._targeted_failing_test_list

    @classmethod
    def from_file(cls, file_path):
        try:
            with open(file_path, "r") as file:
                content = file.read()
            list_content = content.strip().split("\n")
            list_content_striped = [x.strip() for x in list_content]
            list_content_striped_no_comments = list(
                filter(lambda x: not x.startswith("#"), list_content_striped)
            )
            if len(list_content_striped_no_comments) == 0 or all(
                [x == "" for x in list_content_striped_no_comments]
            ):
                return None
            else:
                return TargetFailingTests(list_content_striped_no_comments)
        except FileNotFoundError:
            return None

    @classmethod
    def from_list_string(cls, failing_list):
        failing_list_striped = [x.strip() for x in failing_list]
        if len(failing_list_striped) == 0:
            return None
        else:
            return TargetFailingTests(failing_list_striped)

    def is_target_test(self, current_test_path, current_test_method_name):
        # currentTestFullPath = "::".join([currentTestPath, currentTestMethodName])
        current_test_full_path_generalized = naming_lib.get_generalized_test_name(
            current_test_path, current_test_method_name
        )
        for targetTest in self._targeted_failing_test_list:
            # reformattedTargetFilePath = _reformatPytestTest(targetTest)
            if current_test_full_path_generalized == targetTest:
                return True
        return False



import sqlite3
from pathlib import Path
from typing import List, Tuple


class SbflDbManager:
    _File_name = "fauxpy.db"
    _Test_case_table = "TestCase"
    _Execution_trace_table = "ExecutionTrace"
    _Empty_test_case_table = "EmptyTest"
    _Score_table = "Score"
    _Execution_trace_with_test_type_view = "ExecutionTraceWithTestType"

    def __init__(self, report_directory_path: Path):
        db_file_path = report_directory_path / self._File_name
        db_file_path_str = str(db_file_path.absolute().resolve())
        self._connection = sqlite3.connect(db_file_path_str)
        command_list = self._get_database_schema()
        for command in command_list:
            self._connection.executescript(command)

    def end(self):
        self._connection.close()

    def _get_database_schema(self):
        test_case_table_create_command = (
            f"CREATE TABLE {self._Test_case_table} "
            f"(Rowid INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, "
            f"TestName TEXT NOT NULL UNIQUE, "
            f"Type TEXT NOT NULL,"
            f"Target INTEGER NOT NULL);"
        )
        execution_trace_table_create_command = (
            f"CREATE TABLE {self._Execution_trace_table} "
            f"(Rowid INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, "
            f"TestName TEXT NOT NULL, "
            f"Entity TEXT NOT NULL, "
            f"UNIQUE(TestName, Entity));"
        )
        empty_test_table_create_command = (
            f"CREATE TABLE {self._Empty_test_case_table} "
            f"(Rowid INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, "
            f"TestName TEXT NOT NULL UNIQUE);"
        )

        score_table_create_command = (
            f"CREATE TABLE {self._Score_table} "
            f"(Rowid INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, "
            f"Entity TEXT NOT NULL UNIQUE, "
            f"Ef REAL NOT NULL, "
            f"Ep REAL NOT NULL, "
            f"Nf REAL NOT NULL, "
            f"Np REAL NOT NULL, "
            f"Tarantula REAL NOT NULL, "
            f"Ochiai REAL NOT NULL, "
            f"Dstar REAL NOT NULL);"
        )

        execution_trace_table_index_command = (
            f"CREATE INDEX Index_Entity ON {self._Execution_trace_table} (Entity);"
        )

        test_case_table_index_command = (
            f"CREATE INDEX Index_Type ON {self._Test_case_table} (Type);"
        )

        score_tarantula_table_index_command = (
            f"CREATE INDEX index_Tarantula ON {self._Score_table} (Tarantula);"
        )

        score_ochiai_table_index_command = (
            f"CREATE INDEX index_Ochiai ON {self._Score_table} (Ochiai);"
        )

        score_dstar_table_index_command = (
            f"CREATE INDEX index_Dstar ON {self._Score_table} (Dstar);"
        )

        view_create_command = (
            f"CREATE VIEW {self._Execution_trace_with_test_type_view} AS "
            f"SELECT {self._Execution_trace_table}.Entity, {self._Execution_trace_table}.TestName, {self._Test_case_table}.Type, {self._Test_case_table}.Target "
            f"FROM {self._Execution_trace_table} "
            f"INNER JOIN {self._Test_case_table} ON {self._Execution_trace_table}.TestName = {self._Test_case_table}.TestName"
        )

        command_list = [
            test_case_table_create_command,
            execution_trace_table_create_command,
            empty_test_table_create_command,
            score_table_create_command,
            execution_trace_table_index_command,
            test_case_table_index_command,
            score_tarantula_table_index_command,
            score_ochiai_table_index_command,
            score_dstar_table_index_command,
            view_create_command,
        ]

        return command_list

    def insert_execution_trace(self, test_name: str, covered_entity: List[str]):
        for covEntity in covered_entity:
            cur = self._connection.cursor()
            cur.execute(
                f"INSERT INTO {self._Execution_trace_table} VALUES (NULL, ?, ?)",
                (test_name, covEntity),
            )

        self._connection.commit()

    def insert_empty_test(self, test_name):
        cur = self._connection.cursor()
        cur.execute(
            f"INSERT INTO {self._Empty_test_case_table} VALUES (NULL, ?)", (test_name,)
        )

        self._connection.commit()

    def insert_test_case(self, test_name, test_type, target: bool):
        target_field = 1 if target else 0

        cur = self._connection.cursor()
        cur.execute(
            f"INSERT INTO {self._Test_case_table} VALUES (NULL, ?, ?, ?)",
            (test_name, test_type, target_field),
        )

        self._connection.commit()

    def select_distinct_executed_entity_list(self):
        cur = self._connection.cursor()
        cur.execute(f"SELECT DISTINCT Entity FROM {self._Execution_trace_table}")

        entity_list = [x[0] for x in cur.fetchall()]
        return entity_list

    def select_number_of_tests(self) -> Tuple[int, int]:
        cur = self._connection.cursor()
        cur.execute(
            f"SELECT COUNT(Type) FROM {self._Test_case_table} WHERE Type = 'passed'"
        )
        num_all_passed = cur.fetchone()[0]
        cur.execute(
            f"SELECT COUNT(Type) FROM {self._Test_case_table} WHERE Type = 'failed' AND Target = 1"
        )
        num_all_failed = cur.fetchone()[0]
        # cur.execute(f"SELECT COUNT(*) FROM {_Names.testCaseTable}")
        # numAll = cur.fetchone()[0]
        # assert num_all_passed + num_all_failed == numAll  # Simple checking. Can be removed.

        return num_all_passed, num_all_failed

    def select_number_of_covering_tests(self, entity):
        cur = self._connection.cursor()
        cur.execute(
            f"SELECT COUNT(*) FROM {self._Execution_trace_with_test_type_view} "
            f"WHERE Type = 'passed' AND Entity = ?",
            (entity,),
        )
        num_cov_passed = cur.fetchone()[0]
        cur.execute(
            f"SELECT COUNT(*) FROM {self._Execution_trace_with_test_type_view} "
            f"WHERE Type = 'failed' AND Target = 1 AND Entity = ?",
            (entity,),
        )
        num_cov_failed = cur.fetchone()[0]
        # cur.execute(f"SELECT COUNT(*) FROM {_Names.executionTraceWithTestTypeView} "
        #             f"WHERE Entity = ?", (entity,))
        # numCovAll = cur.fetchone()[0]
        # assert num_cov_passed + num_cov_failed == numCovAll  # Check if left join is causing any NULLs on test type of the view

        return num_cov_passed, num_cov_failed

    def inert_scores(self, entity, ef, ep, nf, np, scores):
        cur = self._connection.cursor()
        cur.execute(
            f"INSERT INTO {self._Score_table} "
            f"VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)",
            (
                entity,
                ef,
                ep,
                nf,
                np,
                scores["Tarantula"],
                scores["Ochiai"],
                scores["Dstar"],
            ),
        )

        self._connection.commit()

    def select_top_n_ranked_entities(self, top_n):
        cur = self._connection.cursor()
        cur.execute(
            f"SELECT Entity, Tarantula FROM {self._Score_table} ORDER BY Tarantula DESC LIMIT ?",
            (top_n,),
        )
        score_tarantula = cur.fetchall()

        cur.execute(
            f"SELECT Entity, Ochiai FROM {self._Score_table} ORDER BY Ochiai DESC LIMIT ?",
            (top_n,),
        )
        score_ochiai = cur.fetchall()

        cur.execute(
            f"SELECT Entity, Dstar FROM {self._Score_table} ORDER BY Dstar DESC LIMIT ?",
            (top_n,),
        )
        score_dstar = cur.fetchall()

        ranked_entities = {
            "Tarantula": score_tarantula,
            "Ochiai": score_ochiai,
            "Dstar": score_dstar,
        }

        return ranked_entities

    def select_all_ranked_entities(self):
        cur = self._connection.cursor()
        cur.execute(
            f"SELECT Entity, Tarantula FROM {self._Score_table} ORDER BY Tarantula DESC"
        )
        score_tarantula = cur.fetchall()

        cur.execute(
            f"SELECT Entity, Ochiai FROM {self._Score_table} ORDER BY Ochiai DESC"
        )
        score_ochiai = cur.fetchall()

        cur.execute(
            f"SELECT Entity, Dstar FROM {self._Score_table} ORDER BY Dstar DESC"
        )
        score_dstar = cur.fetchall()

        ranked_entities = {
            "Tarantula": score_tarantula,
            "Ochiai": score_ochiai,
            "Dstar": score_dstar,
        }

        return ranked_entities



from pathlib import Path
from typing import List

import pytest

from fauxpy.fault_localization.collect_mbfl.session_lib import CollectMbflSession
from fauxpy.fault_localization.collect_ps_info.session_lib import CollectPsInfoSession
from fauxpy.fault_localization.collect_ps_run.session_lib import CollectPsRunSession
from fauxpy.fault_localization.mbfl.session_lib import MbflSession
from fauxpy.fault_localization.ps.session_lib import PsSession
from fauxpy.fault_localization.sbfl.session_lib import SbflSession
from fauxpy.fault_localization.st.session_lib import StSession
from fauxpy.session_lib.fauxpy_printer import fl_print
from fauxpy.session_lib.fl_file_manager import FlFileManager
from fauxpy.session_lib.fl_path_manager import FlPathManager
from fauxpy.session_lib.fl_session import FlSession
from fauxpy.session_lib.fl_type import FlFamily, FlGranularity, MutationStrategy
from fauxpy.session_lib.fauxpy_path import FauxpyPath
from fauxpy.session_lib.targeted_failing_tst import TargetedFailingTst


class FlOptionManager:
    """
    Manages fault localization options and session creation.

    This class handles the parsing and validation of command-line options
    related to fault localization. It also creates and configures FauxPy
    session objects based on the provided options.
    """

    def __init__(
        self,
        project_working_directory: str,
        target_src_opt: str,
        exclude_list_opt: str,
        fl_family_opt: str,
        mutation_strategy_opt: str,
        fl_granularity_opt: str,
        top_n_opt: str,
        failing_file_opt: str,
        failing_list_opt: str,
        fauxpy_verbose_opt: bool,
        file_or_dir,
    ):
        """
        Initializes the FlOptionManager with command-line options.

        Args:
            project_working_directory (str): The project working directory.
            target_src_opt (str): The source directory for fault localization.
            exclude_list_opt (str): Comma-separated list of paths to exclude.
            fl_family_opt (str): Fault localization family option (e.g., sbfl, mbfl).
            mutation_strategy_opt (str): Mutation strategy for Mutation-Based Fault Localization (MBFL).
            fl_granularity_opt (str): Fault localization granularity option (e.g., statement).
            top_n_opt (str): Top N results to consider, or -1 for all results.
            failing_file_opt (str): Path to the file containing targeted failing tests.
            failing_list_opt (str): Comma-separated list of targeted failing tests.
            fauxpy_verbose_opt (bool): Enables detailed output during execution when set to True.
            file_or_dir: Additional file or directory option.
        """
        self._project_working_directory = FauxpyPath.from_relative_path(project_working_directory, ".")
        self._target_src = self._get_validate_target_src(target_src_opt)
        self._exclude_list = self._get_validate_exclude_list(exclude_list_opt)
        self._fl_family = self._get_validate_fl_family(fl_family_opt)
        self._mutation_strategy = self._get_validate_mutation_strategy(mutation_strategy_opt)
        self._fl_granularity = self._get_validate_fl_granularity(fl_granularity_opt)
        self._top_n = self._get_validate_top_n(top_n_opt)
        self._targeted_failing_test_list = (
            self._get_validate_targeted_failing_test_list(
                failing_file_opt, failing_list_opt
            )
        )
        self._file_or_dir_opt = file_or_dir
        self._path_manager = FlPathManager(
            self._project_working_directory,
            self._fl_family,
            self._fl_granularity,
            self._mutation_strategy
        )
        self._fl_file_manager = None

        self._fauxpy_verbose = fauxpy_verbose_opt
        fl_print.set_is_detailed(self._fauxpy_verbose)

    def get_target_src(self) -> FauxpyPath:
        """
        Returns the target source directory.

        Returns:
            FauxpyPath: The validated target source directory.
        """
        return self._target_src

    def get_exclude_list(self) -> List[FauxpyPath]:
        """
        Returns the list of excluded paths.

        Returns:
            List[FauxpyPath]: The validated list of excluded paths.
        """
        return self._exclude_list

    def get_top_n(self) -> int:
        """
        Returns the top N results to consider.

        Returns:
            int: The validated top N value.
        """
        return self._top_n

    def get_fl_family(self) -> FlFamily:
        """
        Returns the fault localization family.

        Returns:
            FlFamily: The fault localization family option.
        """
        return self._fl_family

    def get_mutation_strategy(self) -> MutationStrategy:
        """
        Returns the mutation strategy.

        Returns:
            MutationStrategy: The mutation strategy option.
        """
        return self._mutation_strategy

    def get_fl_granularity(self) -> FlGranularity:
        """
        Returns the fault localization granularity.

        Returns:
            FlGranularity: The fault localization granularity option.
        """
        return self._fl_granularity

    def get_targeted_failing_test_list(self) -> List[TargetedFailingTst]:
        """
        Returns the list of targeted failing tests.

        Returns:
            List[TargetedFailingTst]: The list of targeted failing tests.
        """
        return self._targeted_failing_test_list

    def get_session_file_manager(self):
        """
        Returns the file manager for the fault localization session.

        If the file manager is not yet created, it initializes it.

        Returns:
            FlFileManager: The fault localization file manager.
        """
        if self._fl_file_manager is None:
            report_directory_path = self._path_manager.get_report_directory_path()
            self._fl_file_manager = FlFileManager(report_directory_path)
        return self._fl_file_manager

    def get_fl_session(self) -> FlSession:
        """
        Creates and returns a fault localization session based on the family option.

        Returns:
            FlSession: The created fault localization session.
        """
        report_directory_path = self._path_manager.get_report_directory_path()

        fauxpy_session = None
        if self._fl_family == FlFamily.Sbfl:
            fauxpy_session = SbflSession(
                self._target_src,
                self._exclude_list,
                self._fl_granularity,
                self._top_n,
                self._targeted_failing_test_list,
                report_directory_path,
                Path(self._project_working_directory.get_absolute()),
            )
        elif self._fl_family == FlFamily.St:
            fauxpy_session = StSession(
                self._target_src,
                self._exclude_list,
                FlGranularity.Function,  # ST operates only at the function level granularity.
                self._top_n,
                self._targeted_failing_test_list,
                report_directory_path,
                Path(self._project_working_directory.get_absolute()),
            )
        elif self._fl_family == FlFamily.Mbfl:
            fauxpy_session = MbflSession(
                self._target_src,
                self._exclude_list,
                self._mutation_strategy,
                self._fl_granularity,
                self._top_n,
                self._targeted_failing_test_list,
                self._file_or_dir_opt,
                report_directory_path,
                Path(self._project_working_directory.get_absolute()),
            )
        elif self._fl_family == FlFamily.Ps:
            fauxpy_session = PsSession(
                self._target_src,
                self._exclude_list,
                self._fl_granularity,
                self._top_n,
                self._targeted_failing_test_list,
                report_directory_path,
                Path(self._project_working_directory.get_absolute()),
            )
        elif self._fl_family == FlFamily.CollectMbfl:
            fauxpy_session = CollectMbflSession(
                report_directory_path, Path(self._project_working_directory.get_absolute())
            )
        elif self._fl_family == FlFamily.CollectPsInfo:
            fauxpy_session = CollectPsInfoSession(
                report_directory_path, Path(self._project_working_directory.get_absolute())
            )
        elif self._fl_family == FlFamily.CollectPsRun:
            fauxpy_session = CollectPsRunSession(
                report_directory_path, Path(self._project_working_directory.get_absolute())
            )

        assert fauxpy_session is not None

        return fauxpy_session

    def _get_validate_target_src(self, target_src_opt: str) -> FauxpyPath:
        """
        Validates the target source directory option.

        Args:
            target_src_opt (str): The path to validate.

        Returns:
            FauxpyPath: The validated target source directory.

        Raises:
            pytest.UsageError: If the path does not exist.
        """
        target_src_path_item = FauxpyPath.from_relative_path(
            self._project_working_directory.get_absolute(), target_src_opt
        )

        error_message = (
            f"{target_src_opt} is not a valid option. "
            f"Path {target_src_path_item} does not exist."
        )

        if not target_src_path_item.exists():
            raise pytest.UsageError(error_message)

        return target_src_path_item

    def _get_validate_exclude_list(self, exclude_list_opt: str) -> List[FauxpyPath]:
        """
        Validates the exclude list option.

        Args:
            exclude_list_opt (str): Comma-separated list of paths to exclude.

        Returns:
            List[FauxpyPath]: The validated list of excluded paths.

        Raises:
            pytest.UsageError: If any path does not exist.
        """
        exclude_str_list = self._comma_string_list_to_python_list(exclude_list_opt)

        exclude_path_item_list = []
        for exclude_str_item in exclude_str_list:
            exclude_path_item = FauxpyPath.from_relative_path(
                self._project_working_directory.get_absolute(), exclude_str_item
            )
            if not exclude_path_item.exists():
                error_message = (
                    f"{exclude_list_opt} is not a valid option. "
                    f"Path {exclude_path_item} does not exist."
                )
                raise pytest.UsageError(error_message)
            exclude_path_item_list.append(exclude_path_item)

        return exclude_path_item_list

    @staticmethod
    def _get_validate_fl_family(fl_family_opt: str) -> FlFamily:
        """
        Validates the fault localization family option.

        Args:
            fl_family_opt (str): The fault localization family option.

        Returns:
            FlFamily: The validated fault localization family.

        Raises:
            pytest.UsageError: If the option is not valid.
        """
        error_message = (
            f"{fl_family_opt} is not a valid option. "
            f"Correct options: sbfl, mbfl, ps, st."
        )

        if fl_family_opt not in [
            "sbfl",
            "st",
            "mbfl",
            "ps",
            "collectmbfl",
            "collectpsinfo",
            "collectpsrun",
        ]:
            raise pytest.UsageError(error_message)

        fl_family = FlFamily.Sbfl
        if fl_family_opt == "mbfl":
            fl_family = FlFamily.Mbfl
        elif fl_family_opt == "ps":
            fl_family = FlFamily.Ps
        elif fl_family_opt == "st":
            fl_family = FlFamily.St
        elif fl_family_opt == "collectmbfl":
            fl_family = FlFamily.CollectMbfl
        elif fl_family_opt == "collectpsinfo":
            fl_family = FlFamily.CollectPsInfo
        elif fl_family_opt == "collectpsrun":
            fl_family = FlFamily.CollectPsRun

        return fl_family

    @staticmethod
    def _get_validate_mutation_strategy(mutation_strategy_opt: str) -> MutationStrategy:
        """
        Validates the mutation strategy option.

        Args:
            mutation_strategy_opt (str): The mutation strategy option.

        Returns:
            MutationStrategy: The validated mutation strategy.

        Raises:
            pytest.UsageError: If the option is not valid.
        """
        error_message = (
            f"{mutation_strategy_opt} is not a valid option. "
            f"Valid options: t, tgpt4ominiapi, gpt4ominiapi, tgpt4oapi, gpt4oapi."
        )

        if mutation_strategy_opt not in ["t", "tgpt4ominiapi", "gpt4ominiapi", "tgpt4oapi", "gpt4oapi"]:
            raise pytest.UsageError(error_message)

        mutation_strategy = MutationStrategy.Traditional
        if mutation_strategy_opt == "tgpt4ominiapi":
            mutation_strategy = MutationStrategy.TraditionalWithGPT4oMini
        elif mutation_strategy_opt == "gpt4ominiapi":
            mutation_strategy = MutationStrategy.GPT4oMini
        elif mutation_strategy_opt == "tgpt4oapi":
            mutation_strategy = MutationStrategy.TraditionalWithGPT4o
        elif mutation_strategy_opt == "gpt4oapi":
            mutation_strategy = MutationStrategy.GPT4o

        return mutation_strategy

    @staticmethod
    def _get_validate_fl_granularity(fl_granularity_opt: str) -> FlGranularity:
        """
        Validates the fault localization granularity option.

        Args:
            fl_granularity_opt (str): The fault localization granularity option
                                      (e.g., statement, function, s, f).

        Returns:
            FlGranularity: The validated fault localization granularity.

        Raises:
            pytest.UsageError: If the option is not valid.
        """
        error_message = (
            f"{fl_granularity_opt} is not a valid option. "
            f"Valid options: statement (or s), function (or f)."
        )

        granularity_map = {
            "statement": FlGranularity.Statement,
            "s": FlGranularity.Statement,
            "function": FlGranularity.Function,
            "f": FlGranularity.Function,
        }

        if fl_granularity_opt not in granularity_map:
            raise pytest.UsageError(error_message)

        return granularity_map[fl_granularity_opt]

    @staticmethod
    def _get_validate_top_n(top_n_opt: str) -> int:
        """
        Validates the top N results option.

        Args:
            top_n_opt (str): The top N option, which should be -1 or an integer >= 1.

        Returns:
            int: The validated top N value.

        Raises:
            pytest.UsageError: If the option is not a valid integer or is less than -1 or zero.
        """
        error_message = (
            f"{top_n_opt} is not a valid option. "
            f"Correct options: -1 or an integer x, where x >= 1."
        )
        try:
            int_value = int(top_n_opt)
        except:
            raise pytest.UsageError(error_message)

        if int_value < -1 or int_value == 0:
            raise pytest.UsageError(error_message)

        return int_value

    @staticmethod
    def _comma_string_list_to_python_list(comma_string_list: str) -> List[str]:
        """
        Converts a comma-separated string list to a Python list of strings.

        Args:
            comma_string_list (str): The comma-separated string list to convert.

        Returns:
            List[str]: The list of strings obtained from the input string.
        """
        stripped_string = comma_string_list.strip()
        brackets_removed = stripped_string[1:-1]
        if brackets_removed == "":
            return []
        else:
            items = brackets_removed.split(",")
            stripped_item_list = [x.strip() for x in items]

            return stripped_item_list

    @staticmethod
    def _multi_line_string_list_to_python_list(
        multi_line_string_list: str,
    ) -> List[str]:
        """
        Converts a multi-line string list to a Python list of strings.

        Args:
            multi_line_string_list (str): The multi-line string list to convert.

        Returns:
            List[str]: The list of strings obtained from the input string, excluding comments and empty lines.
        """
        list_content = multi_line_string_list.strip().split("\n")
        list_content_striped = [x.strip() for x in list_content]
        list_content_striped_no_comments = list(
            filter(lambda x: not x.startswith("#"), list_content_striped)
        )
        if len(list_content_striped_no_comments) == 0 or all(
            [x == "" for x in list_content_striped_no_comments]
        ):
            return []
        else:
            return list_content_striped_no_comments

    def _get_validate_targeted_failing_test_list(
        self, failing_file_opt: str, failing_list_opt: str
    ) -> List[TargetedFailingTst]:
        """
        Validates and returns the list of targeted failing tests based on the provided options.

        Args:
            failing_file_opt (str): Path to the file containing targeted failing tests (optional).
            failing_list_opt (str): Comma-separated list of targeted failing tests (optional).

        Returns:
            List[TargetedFailingTst]: The validated list of targeted failing tests.

        Raises:
            pytest.UsageError: If both file and list options are specified, or if any test item or path is invalid.
        """
        if failing_file_opt is not None and failing_list_opt is not None:
            error_message = (
                "You may specify either "
                "--failing-list or --failing-file,"
                " but not both."
            )
            raise pytest.UsageError(error_message)

        targeted_failing_test_string_list = []
        if failing_list_opt is not None:
            targeted_failing_test_string_list = self._comma_string_list_to_python_list(
                failing_list_opt
            )
        elif failing_file_opt is not None:
            failing_file_path = Path(self._project_working_directory.get_absolute()) / failing_file_opt
            if not failing_file_path.exists():
                error_message = (
                    f"{failing_file_opt} is not a valid option. "
                    f"Path {failing_file_path} does not exist."
                )
                raise pytest.UsageError(error_message)
            failing_file_content = failing_file_path.read_text()
            targeted_failing_test_string_list = (
                self._multi_line_string_list_to_python_list(failing_file_content)
            )

        targeted_failing_test_item_list = []
        for targeted_failing_test_string_item in targeted_failing_test_string_list:
            target_failing_test_item = self._get_validate_targeted_failing_test_item(
                targeted_failing_test_string_item
            )
            targeted_failing_test_item_list.append(target_failing_test_item)
        return targeted_failing_test_item_list

    def _get_validate_targeted_failing_test_item(
        self, targeted_failing_test_string_item
    ) -> TargetedFailingTst:
        """
        Validates and returns a targeted failing test item based on the provided string.

        Args:
            targeted_failing_test_string_item (str): The string representing the targeted failing test.

        Returns:
            TargetedFailingTst: The validated targeted failing test item.

        Raises:
            pytest.UsageError: If the test item format is incorrect or the file containing the test does not exist.
        """
        content_parts = targeted_failing_test_string_item.split("::")
        if len(content_parts) == 2:
            module_path_item = FauxpyPath.from_relative_path(
                self._project_working_directory.get_absolute(), content_parts[0]
            )
            if not module_path_item.exists():
                error_message = (
                    f"{targeted_failing_test_string_item} is not a valid option. "
                    f"Path {module_path_item} does not exist."
                )
                raise pytest.UsageError(error_message)
            function_name = content_parts[1]
            targeted_failing_test_item = TargetedFailingTst(
                module_path_item, None, function_name
            )
        elif len(content_parts) == 3:
            module_path_item = FauxpyPath.from_relative_path(
                self._project_working_directory.get_absolute(), content_parts[0]
            )
            if not module_path_item.exists():
                error_message = (
                    f"{targeted_failing_test_string_item} is not a valid option. "
                    f"Path {module_path_item} does not exist."
                )
                raise pytest.UsageError(error_message)
            class_name = content_parts[1]
            function_name = content_parts[2]
            targeted_failing_test_item = TargetedFailingTst(
                module_path_item, class_name, function_name
            )
        else:
            error_message = (
                f"{targeted_failing_test_string_item} is not a valid option. "
                f"The format of the test is incorrect."
            )
            raise pytest.UsageError(error_message)

        return targeted_failing_test_item



import logging
from typing import Optional

from fauxpy.command_line.pytest_mode.pytest_option_manager import PytestOptionManager
from fauxpy.session_lib.fauxpy_printer import fl_print
from fauxpy.session_lib.fauxpy_session_type import FauxpySessionType
from fauxpy.session_lib.fl_family_session import FlFamilySession
from fauxpy.session_lib.fl_session import FlSession
from fauxpy.session_lib.fl_session_report import FlSessionReport
from fauxpy.session_lib.timer import Timer


class FauxpyPytestModeHandler:
    """
    Handles Pytest Mode by interacting with Pytest through various hooks.
    """

    def __init__(self):
        self._session_timer = Timer()
        self._pytest_option_manager = PytestOptionManager()
        self._fauxpy_session_type = FauxpySessionType.FauxpyNotCalled
        self._session_object: Optional[FlSession] = None
        self._session_file_manager = None

    def add_option(self, pytest_parser):
        """
        Adds FauxPy-specific command-line options to the pytest parser.

        Args:
            pytest_parser: The Pytest parser object to which FauxPy options should be added.
        """
        self._pytest_option_manager.add_fauxpy_options(pytest_parser)

    def configure(self, pytest_config):
        """
        Configures the FauxPy session.

        Args:
            pytest_config: The Pytest configuration object used to determine the session type
            and options.
        """
        self._fauxpy_session_type = self._pytest_option_manager.get_fauxpy_session_type(
            pytest_config
        )
        if self._fauxpy_session_type == FauxpySessionType.FaultLocalization:
            self._session_timer.start()

            fl_option_manager = self._pytest_option_manager.get_fl_option_manager(
                pytest_config
            )

            self._session_object = fl_option_manager.get_fl_session()
            self._session_file_manager = fl_option_manager.get_session_file_manager()

            logging.basicConfig(
                filename=self._session_file_manager.get_log_file_path(),
                level=logging.DEBUG,
                format="%(asctime)s %(message)s",
                filemode="w",
            )

            if isinstance(self._session_object, FlFamilySession):
                self._session_file_manager.save_config_to_file(
                    target_src=fl_option_manager.get_target_src(),
                    exclude_list=fl_option_manager.get_exclude_list(),
                    fl_family=fl_option_manager.get_fl_family(),
                    fl_granularity=fl_option_manager.get_fl_granularity(),
                    top_n=fl_option_manager.get_top_n(),
                    targeted_failing_test_list=fl_option_manager.get_targeted_failing_test_list(),
                    mutation_strategy=fl_option_manager.get_mutation_strategy()
                )

    def runtest_call(self, item):
        """
        Handles the test call event during test execution.

        Args:
            item: The pytest item representing the test being executed.
        """
        if self._fauxpy_session_type == FauxpySessionType.FaultLocalization:
            self._session_object.run_test_call(item)

    def runtest_make_report(self, item, call):
        """
        Gathers the test results and generates reports after a test call is executed.

        Args:
            item: The Pytest item representing the test that was executed.
            call: The Pytest call object containing the results of the test (setup, call, teardown).
        """
        if self._fauxpy_session_type == FauxpySessionType.FaultLocalization:
            self._session_object.run_test_make_report(item, call)

    def terminal_summary(self, terminal_reporter, exitstatus):
        """
        Generates final results after all tests have completed.

        Args:
            terminal_reporter: The Pytest terminal reporter object used for output.
            exitstatus: The exit status of Pytest, indicating success or failure.
        """
        if self._fauxpy_session_type == FauxpySessionType.FaultLocalization:
            banner_start = (
                "***************************************************\n"
                "                FauxPy Started!                    \n"
                "***************************************************\n"
            )
            print("\n")
            print(banner_start)

            fl_print.normal(f"Running {self._session_object}")

            if (isinstance(self._session_object, FlFamilySession) and
                    len(self._session_object.get_targeted_failing_test_list()) > 0):
                fl_print.normal("Targeted failing tests:")
                for index, item in enumerate(self._session_object.get_targeted_failing_test_list()):
                    fl_print.normal(f"  {index + 1}. {item}")

            print("\n")
            banner_dynamic_analysis = (
                               "==============================\n"
                               " Dynamic Analysis in Progress \n"
                               "==============================\n"
            )
            print(banner_dynamic_analysis)

            score_entity_list = self._session_object.terminal_summary(
                terminal_reporter, exitstatus
            )

            print("\n")
            print("--- Dynamic Analysis Complete ---")

            session_time = self._session_timer.end()
            self._session_file_manager.save_delta_time_to_file(session_time)

            for technique, score_list in score_entity_list.items():
                self._session_file_manager.save_scores_to_file(technique, score_list)

            if isinstance(self._session_object, FlFamilySession):
                fl_session_report = FlSessionReport(
                    score_entity_list,
                    session_time,
                    self._session_object.get_fl_granularity(),
                    self._session_object.get_project_working_directory()
                )
                report_string = fl_session_report.generate_report()
                print(report_string)

            banner_end = (
                "**************************************************\n"
                "                FauxPy Ended!                     \n"
                "**************************************************\n"
            )
            print(banner_end)



import re
from pathlib import Path
from typing import Tuple, List

from fauxpy.fault_localization.util.path_util import PathUtil
from fauxpy.session_lib.fauxpy_path import FauxpyPath


class TracebackParser:
    def __init__(self, project_working_directory: Path):
        self._path_util = PathUtil(project_working_directory)

    @staticmethod
    def _get_required_lines(lines) -> str:
        lines_list = []
        for line in lines:
            if line.startswith("> "):
                lines_list.append(line)
            if line.startswith("E "):
                lines_list.append(line)

        if len(lines_list) == 0:
            lines_list.append(lines[-1])

        information = "::".join(lines_list)
        return information

    def _get_trace_back_entry_essence(self, entry) -> str:
        file_path = entry.reprfileloc.path
        line_number = entry.reprfileloc.lineno
        message = entry.reprfileloc.message
        lines = self._get_required_lines(entry.lines)
        information = "::".join([file_path, str(line_number), message, lines])
        return information

    def get_short_trace_back_info(self, trace_back) -> str:
        information = []
        for item in trace_back.reprentries:
            item_info = self._get_trace_back_entry_essence(item)
            information.append(item_info)

        info_string = "\n".join(information)
        return info_string

    @staticmethod
    def has_timeout_happened(long_repr_text) -> bool:
        # Pattern is "E       Failed: Timeout >"
        pattern = "E( +)Failed:( *)Timeout( *)>"
        mat = re.search(pattern, long_repr_text)
        return mat is not None

    @staticmethod
    def _is_python_module(path: str):
        return path.endswith(".py")

    def get_exception_location(
        self, traceback, target_src: FauxpyPath, exclude_list: List[FauxpyPath]
    ) -> Tuple[str, int]:
        exception_file_path = ""
        exception_line_number = -1
        tb_len = len(traceback.reprentries)
        if tb_len > 1:
            for i in range(len(traceback.reprentries)):
                tmp = tb_len - i - 1
                c_path = traceback.reprentries[tb_len - i - 1].reprfileloc.path
                c_line_number = traceback.reprentries[tb_len - i - 1].reprfileloc.lineno
                c_path_abs = self._path_util.relative_path_to_abs_path(c_path)
                if self._path_util.path_should_be_localized(
                    target_src, exclude_list, c_path_abs
                ) and self._is_python_module(c_path_abs):
                    exception_file_path = c_path
                    exception_line_number = c_line_number
                    break

        return exception_file_path, exception_line_number



from fauxpy.fault_localization.collect_ps_info.api_lib import CollectPsInfoApi


def test_index_predicate_sequence():
    predicate_sequence = "Pred_1,Pred_2,Pred_3,Pred_3,Pred_2,Pred_2,Pred_6"
    collect_ps_info_api = CollectPsInfoApi()
    actual = collect_ps_info_api.index_predicate_sequence(predicate_sequence)
    expected = "Pred_1::0,Pred_2::0,Pred_3::0,Pred_3::1,Pred_2::1,Pred_2::2,Pred_6::0"
    assert actual == expected


def test_get_index_predicate_sequences():
    predicate_sequence_table = [
        [
            "tests/test_black.py::716::BlackTestCase.test_get_future_imports",
            "Pred_1,Pred_2,Pred_3,Pred_3,Pred_2,Pred_2,Pred_6",
        ],
        [
            "tests/test_black.py::716::BlackTestCase.test_get_future_imports1",
            "Pred_1,Pred_2,Pred_4,Pred_5,Pred_6,Pred_7,Pred_8",
        ],
    ]
    collect_ps_info_api = CollectPsInfoApi()
    actual = collect_ps_info_api.get_index_predicate_sequences(predicate_sequence_table)
    expected = [
        (
            "tests/test_black.py::716::BlackTestCase.test_get_future_imports",
            "Pred_1,Pred_2,Pred_3,Pred_3,Pred_2,Pred_2,Pred_6",
            "Pred_1::0,Pred_2::0,Pred_3::0,Pred_3::1,Pred_2::1,Pred_2::2,Pred_6::0",
        ),
        (
            "tests/test_black.py::716::BlackTestCase.test_get_future_imports1",
            "Pred_1,Pred_2,Pred_4,Pred_5,Pred_6,Pred_7,Pred_8",
            "Pred_1::0,Pred_2::0,Pred_4::0,Pred_5::0,Pred_6::0,Pred_7::0,Pred_8::0",
        ),
    ]
    assert actual == expected



import os

import pytest

from fauxpy.session_lib.fauxpy_path import FauxpyPath


def test_path_lib_from_relative_1():
    pwd = "/home/user"
    rel = "a/b/c.py"
    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_relative_2():
    pwd = "/home/user/"
    rel = "a/b/c.py"
    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_relative_3():
    pwd = "/home/user"
    rel = "./a/b/c.py"
    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_relative_4():
    pwd = "/home/user/././"
    rel = "a/b/c.py"
    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_relative_5():
    pwd = "/home/user"
    rel = "./././././a/b/c.py"
    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_relative_6():
    pwd = "/home/user"
    rel = "."
    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "."
    assert fauxpy_path.get_absolute() == "/home/user"

def test_path_lib_from_absolute_1():
    absolute_path = "/home/user/a/b/c.py"
    pwd = "/home/user"
    fauxpy_path = FauxpyPath.from_absolute_path(pwd, absolute_path)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_absolute_2():
    absolute_path = "/home/././././user/a/b/c.py"
    pwd = "/home/user"
    fauxpy_path = FauxpyPath.from_absolute_path(pwd, absolute_path)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_absolute_3():
    absolute_path = "/home/user/a/b/c.py"
    pwd = "/home/user/"
    fauxpy_path = FauxpyPath.from_absolute_path(pwd, absolute_path)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_absolute_4():
    absolute_path = "/home/user/a/b/c.py"
    pwd = "/home/user/./././././."
    fauxpy_path = FauxpyPath.from_absolute_path(pwd, absolute_path)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

def test_path_lib_from_absolute_5():
    absolute_path = "/home/user"
    pwd = "/home/user"
    fauxpy_path = FauxpyPath.from_absolute_path(pwd, absolute_path)
    assert fauxpy_path.get_relative() == "."
    assert fauxpy_path.get_absolute() == "/home/user"

def test_path_lib_changing_cwd():
    pwd = "/home/user"
    rel = "a/b/c.py"
    absolute_path = "/home/user/a/b/c.py"

    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

    fauxpy_path = FauxpyPath.from_absolute_path(pwd, absolute_path)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

    os.chdir("/")

    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"

    fauxpy_path = FauxpyPath.from_absolute_path(pwd, absolute_path)
    assert fauxpy_path.get_relative() == "a/b/c.py"
    assert fauxpy_path.get_absolute() == "/home/user/a/b/c.py"


def test_path_lib_dot_path():
    pwd = "/home/user"
    rel = "."
    fauxpy_path = FauxpyPath.from_relative_path(pwd, rel)
    assert fauxpy_path.get_relative() == "."
    assert fauxpy_path.get_absolute() == "/home/user"

def test_path_lib_using_constructor():
    with pytest.raises(NotImplementedError):
        FauxpyPath()



from fauxpy.fault_localization.mbfl.db_manager import MbflDbManager
from fauxpy.fault_localization.mbfl.metric_metallaxis import MetricMetallaxis
from fauxpy.fault_localization.mbfl.metric_muse import MetricMuse


class EntityScoreManager:
    def __init__(self, db_manager: MbflDbManager):
        self._db_manager = db_manager

    @staticmethod
    def _compute_entity_scores(mutant_score_list):
        muse_mutant_score_list = mutant_score_list["Muse"]
        metallaxis_mutant_scores = mutant_score_list["Metallaxis"]

        muse_entity_score = MetricMuse.compute_entity_score(muse_mutant_score_list)
        metallaxis_entity_score = MetricMetallaxis.compute_entity_score(
            metallaxis_mutant_scores
        )

        return {"Muse": muse_entity_score, "Metallaxis": metallaxis_entity_score}

    def compute_entity_scores_store_db(self, top_n: int):
        entities = self._db_manager.select_distinct_all_mutant_score_terms_entities()

        for entity in entities:
            mutant_scores = self._db_manager.select_mutant_score_terms_scores(entity)
            entity_scores = self._compute_entity_scores(mutant_scores)

            entity_muse_score = entity_scores["Muse"]
            entity_metallaxis_score = entity_scores["Metallaxis"]

            self._db_manager.insert_entity_score(
                entity, entity_muse_score, entity_metallaxis_score
            )

        if top_n == -1:
            score_entities = self._db_manager.select_all_ranked_entities()
        elif top_n >= 1:
            pass
            score_entities = self._db_manager.select_top_n_ranked_entities(top_n)
        else:
            raise Exception(f"TopN {top_n} is not supported.")

        return score_entities



from typing import Tuple


def convert_test_name_to_components(test_name: str) -> Tuple[str, int, str]:
    components = test_name.split("::")
    return components[0], int(components[1]), components[2]


def get_statement_name(path: str, line_number: int):
    statement_name = path + "::" + str(line_number)
    return statement_name


def convert_statement_name_to_components(statement_name: str) -> Tuple[str, int]:
    components = statement_name.split("::")
    return components[0], int(components[1])


def get_covered_function_name(
    path: str, function_name: str, line_start: int, line_end: int
):
    cov_func_name = (
        path + "::" + function_name + "::" + str(line_start) + "::" + str(line_end)
    )
    return cov_func_name


def test_name_to_file_name(test_name: str):
    new_test_name = test_name.replace("/", "_").replace("/", "_").replace(":", "_")
    return new_test_name


def get_generalized_test_name(file_path: str, function_name: str) -> str:
    """
    For parametrized tests, this function removes the parameter values,
    leaving only the function name to represent all variations of the test.
    For non-parametrized tests, the original test name is returned as is.
    """
    # filePath, _, functionName = convertTestNameToComponents(testName)
    generalized_function_name = function_name.split("[")[0]

    # For CLASS_NAME.FUNCTION_NAME
    if "." in generalized_function_name:
        class_name, function_name = generalized_function_name.split(".")
        generalized_function_name = "::".join([class_name, function_name])

    generalized_test_func_path = "::".join([file_path, generalized_function_name])
    return generalized_test_func_path



import ast
from typing import Any

import astor


class InstrumentationTransformer(ast.NodeTransformer):
    def __init__(self, tree, candidate_predicates, seen_exceptions):
        self.tree = tree
        self.candidatePredicates = candidate_predicates
        self.seenExceptions = seen_exceptions

    def is_test_node(self, node: ast.AST) -> bool:
        class TestVisitor(ast.NodeVisitor):
            def __init__(self, nd):
                self._node = nd
                self._is_test_node = False

            def visit(self, n: ast.AST) -> Any:
                if self._is_test_node:
                    return

                if hasattr(n, "test"):
                    if self._node == n.test:
                        self._is_test_node = True

                self.generic_visit(n)

            def is_test_node(self):
                return self._is_test_node

        test_visitor = TestVisitor(node)
        test_visitor.visit(self.tree)
        is_test_node = test_visitor.is_test_node()
        return is_test_node

    def visit(self, node: ast.AST) -> Any:
        candidate_name = self._get_candidate_name(node)
        seen_exp_name = self._get_seen_exception_name(node)

        if candidate_name is not None and self.is_test_node(node):
            return self._instrument_test_predicate(node, candidate_name)
        if seen_exp_name is not None:
            return self._instrument_seen_exception(node, seen_exp_name)

        return self.generic_visit(node)

    # TODO: if node.end_lineno exists, use it. Otherwise, the following one.
    def _get_ending_line(self, node):
        line_end_max = -1

        if hasattr(node, "__dict__"):
            all_att = node.__dict__
            for att in all_att:
                att_instance = node.__getattribute__(att)
                if isinstance(att_instance, list):
                    for subAtt in att_instance:
                        line_end_max = max(
                            self._get_ending_line(subAtt),
                            line_end_max,
                        )
                else:
                    line_end_max = max(
                        self._get_ending_line(att_instance),
                        line_end_max,
                    )
                if hasattr(att_instance, "lineno"):
                    line_end_max = max(att_instance.lineno, line_end_max)

        if hasattr(node, "lineno"):
            line_end_max = max(node.lineno, line_end_max)
        return line_end_max

    @staticmethod
    def _instrument_test_predicate(node, candidate_name):
        node_as_text = astor.to_source(node).strip()
        new_node_as_text = (
            f"fauxpy_inst.wrap_pred_to_switch({node_as_text}, '{candidate_name}')"
        )
        new_ast = ast.parse(new_node_as_text)
        new_node_ast = new_ast.body[0].value
        return new_node_ast

    @staticmethod
    def _instrument_seen_exception(node, seen_exp_name):
        # TODO: the solution is ad hoc. Find a better way for it.
        node_as_text = astor.to_source(node).strip()
        visit_exp_statement_as_text = (
            f"fauxpy_inst.exception_seen_at_next_line('{seen_exp_name}')"
        )
        new_node_as_text = f"{visit_exp_statement_as_text}\n{node_as_text}"
        new_node_ast = ast.parse(new_node_as_text)
        return new_node_ast

    def _get_candidate_name(self, node):
        if hasattr(node, "lineno"):
            for predicate in self.candidatePredicates:
                line_start, line_end, candidate_name = predicate
                if line_start == node.lineno and line_end == self._get_ending_line(
                    node
                ):
                    return candidate_name
        return None

    def _get_seen_exception_name(self, node):
        if hasattr(node, "lineno"):
            for seenException in self.seenExceptions:
                line_number, seen_exp_name = seenException
                if line_number == node.lineno:
                    return seen_exp_name
        return None



# Constants for PyLLMut
MUTANTS_PER_LINE_COUNT = 7  # Number of mutants generated per line
TIMEOUT_SECONDS_PER_LINE = 10  # Timeout in seconds per line during mutation generation

# Constants representing pytest phases and outcomes
PYTEST_CALL = "call"  # Represents the "call" phase of a test in pytest
PYTEST_PASSED = "passed"  # Indicates a test that has passed
PYTEST_FAILED = "failed"  # Indicates a test that has failed

# Constants for FauxPy fault localization granularity levels
FAUXPY_STATEMENT = (
    "statement"  # Specifies statement-level granularity in fault localization
)
FAUXPY_FUNCTION = (
    "function"  # Specifies function-level granularity in fault localization
)

# Fl session constants
DB_FILE_NAME_FL_SESSION = "fauxpy.db"  # Database file name for FauxPy fault localization session
CONFIG_FILE_NAME_FL_SESSION = "config.json"  # Configuration file name for FauxPy fault localization session
TIME_FILE_NAME_FL_SESSION = "deltaTime.json"  # Time-related file for FauxPy fault localization session
LOG_FILE_NAME_FL_SESSION = "logging.log"  # Log file name for FauxPy fault localization session
REPORT_DIRECTORY_NAME_PREFIX_FL_SESSION = "FauxPyReport"  # Prefix for the directory name including FauxPy fault localization reports

# Fl results constants
SCORES_CSV_HEADER = ["Entity", "Score"]

class FileNames(object):
    CollectModeDirectoryName = "FauxPyCollectModeReport"
    collectModeTestCases = "collectModeTestCases.json"
    collectModeCoveredLinesForTest = "collectModeCoveredLinesForTest.json"
    collectModePredicateSequences = "collectModePredicateSequences.json"
    collectModeSeenExceptions = "collectModeSeenExceptions.json"
    instrumentationCollectModeExecutedPredicateSequenceFileName = (
        "instrumentationCollectModeExecutedPredicateSequence.txt"
    )
    instrumentationCollectModeConfigFileName = "instrumentationCollectModeConfig.txt"
    instrumentationCollectModeEvaluationCounterFileName = (
        "instrumentationCollectModeEvaluationCounter.txt"
    )
    instrumentationCollectModeExceptionSeenFileName = (
        "instrumentationCollectModeExceptionSeen.txt"
    )


def getCollectModeExecutedPredicateSequenceFileName():
    return FileNames.instrumentationCollectModeExecutedPredicateSequenceFileName


def getCollectModeConfigFileName():
    return FileNames.instrumentationCollectModeConfigFileName


def getCollectModeEvaluationCounterFileName():
    return FileNames.instrumentationCollectModeEvaluationCounterFileName


def getExceptionSeenFileName():
    return FileNames.instrumentationCollectModeExceptionSeenFileName


# Constants used to compute the timeout window for
# running mutants in MBFL and during predicate switching
# in PS
# TODO: Maybe finding better values for these two is a good idea.
TEST_TIMEOUT_SCALING_FACTOR = 2
TEST_TIMEOUT_OFFSET = 5



import sqlite3
from pathlib import Path


class StDbManager:
    _File_name = "fauxpy.db"
    _Traceback_score_table = "TracebackScore"

    def __init__(self, report_directory_path: Path):
        db_file_path = report_directory_path / self._File_name
        db_file_path_str = str(db_file_path.absolute().resolve())
        self._connection = sqlite3.connect(db_file_path_str)
        command_list = self._get_database_schema()
        for command in command_list:
            self._connection.executescript(command)

    def _get_database_schema(self):
        traceback_score_table_create_command = (
            f"CREATE TABLE {self._Traceback_score_table} "
            f"(Rowid INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, "
            f"TestName TEXT NOT NULL, "
            f"Function TEXT NOT NULL,"
            f"Score REAL NOT NULL);"
        )

        traceback_score_table_index_command = f"CREATE INDEX Index_FunctionName ON {self._Traceback_score_table} (Function);"

        command_list = [
            traceback_score_table_create_command,
            traceback_score_table_index_command,
        ]

        return command_list

    def end(self):
        self._connection.close()

    def insert_traceback_scores(self, test_name, test_scores):
        for testScore in test_scores:
            cur = self._connection.cursor()
            cur.execute(
                f"INSERT INTO {self._Traceback_score_table} VALUES (NULL, ?, ?, ?)",
                (test_name, testScore[0], testScore[1]),
            )

        self._connection.commit()

    def select_all_ranked_functions(self):
        cur = self._connection.cursor()
        cur.execute(
            f"SELECT Function, MAX(Score) FROM {self._Traceback_score_table} "
            f"GROUP BY Function "
            f"ORDER BY Score DESC"
        )

        return cur.fetchall()

    def select_top_n_ranked_functions(self, top_n):
        cur = self._connection.cursor()
        cur.execute(
            f"SELECT Function, MAX(Score) FROM {self._Traceback_score_table} "
            f"GROUP BY Function "
            f"ORDER BY Score DESC "
            f"LIMIT ?",
            (top_n,),
        )

        return cur.fetchall()



from pathlib import Path
from typing import List

import coverage

from fauxpy import program_tracer
from fauxpy.command_line.pytest_mode import legacy_input
from fauxpy.fault_localization.granularity.db_manager import FunctionLevelDbManager
from fauxpy.fault_localization.ps.ps_run_manager import PsRunManager
from fauxpy.fault_localization.ps.db_manager import PsDbManager
from fauxpy.fault_localization.util import timeout
from fauxpy.fault_localization.util.path_util import PathUtil
from fauxpy.fault_localization.util.traceback_lib import TracebackParser
from fauxpy.session_lib.fl_family_session import FlFamilySession
from fauxpy.session_lib.fl_type import FlGranularity, FlFamily
from fauxpy.session_lib.fauxpy_path import FauxpyPath
from fauxpy.session_lib.pytest_tst_item import PytestTstItem
from fauxpy.session_lib.timer import Timer
from fauxpy.session_lib.targeted_failing_tst import TargetedFailingTst


class PsSession(FlFamilySession):

    # TODO: Program tracer does not work right now.
    _Use_coverage_lib = True

    def __init__(
        self,
        target_src: FauxpyPath,
        exclude_list: List[FauxpyPath],
        fl_granularity: FlGranularity,
        top_n: int,
        targeted_failing_test_list: List[TargetedFailingTst],
        report_directory_path: Path,
        project_working_directory: Path,
    ):
        self._target_src = target_src
        self._exclude_list = exclude_list
        self._fl_granularity = fl_granularity
        self._top_n = top_n
        self._targeted_failing_test_list = targeted_failing_test_list

        self._coverage_manager = coverage.Coverage()
        self._db_manager = PsDbManager(report_directory_path)
        self._current_test_name = None
        self._current_test_timer = Timer()
        self._function_level_db_manager = FunctionLevelDbManager(report_directory_path)
        self._project_working_directory = project_working_directory
        self._ps_run_manager = PsRunManager(
            self._db_manager, self._function_level_db_manager, project_working_directory
        )
        self._traceback_parser = TracebackParser(project_working_directory)
        self._path_util = PathUtil(project_working_directory)

    def __str__(self):
        return "PS session"

    def get_fl_granularity(self) -> FlGranularity:
        return self._fl_granularity

    def get_fl_family(self) -> FlFamily:
        return FlFamily.Ps

    def get_project_working_directory(self) -> FauxpyPath:
        return FauxpyPath.from_relative_path(str(self._project_working_directory), ".")

    def get_targeted_failing_test_list(self) -> List[TargetedFailingTst]:
        """
        Returns the list of targeted failing tests for fault localization.

        Returns:
            list: A list of targeted failing tests.
        """
        return self._targeted_failing_test_list

    def run_test_call(self, item):
        self._current_test_timer.start()

        self._current_test_name = PytestTstItem(item).get_test_name()

        if self._current_test_name:
            self._coverage_manager.start()
        else:
            program_tracer.start(
                isWanted=lambda x: self._path_util.path_should_be_localized(
                    self._target_src,
                    self._exclude_list,
                    x,
                )
            )

    def run_test_make_report(self, item, call):
        # TODO: Replace custom tracer with coverage library (commented code).

        if call.when == "call":
            test_name = PytestTstItem(item).get_test_name()
            if test_name != self._current_test_name:
                raise Exception(
                    f"Starting coverage for {self._current_test_name}. But closing coverage for {test_name}."
                )

            if self._Use_coverage_lib:
                self._coverage_manager.stop()
                coverage_data = self._coverage_manager.get_data()
                covered_statement_list = []
                covered_file_list = coverage_data.measured_files()
                for file in covered_file_list:
                    if self._path_util.path_should_be_localized(
                        self._target_src,
                        self._exclude_list,
                        file,
                    ):
                        lines = coverage_data.lines(file)
                        for line in lines:
                            covered_statement_list.append((file, line))
                self._coverage_manager.erase()
            else:
                program_tracer.stop()
                execution_trace = program_tracer.getExecutionTrace()
                covered_statement_list = (
                    execution_trace.getExecutedLinesInOrderOfExecution()
                )

            if len(covered_statement_list) == 0:
                self._db_manager.insert_empty_test(test_name)
            else:
                for coveredStmt in covered_statement_list:
                    self._db_manager.insert_covered_line_for_test(
                        test_name, coveredStmt[0], coveredStmt[1]
                    )

            # Let's give more time to mutants by putting these lines after
            current_test_time = self._current_test_timer.end()
            self._db_manager.insert_test_time(test_name, current_test_time)

    def terminal_summary(self, terminal_reporter, exit_status):
        for key, value in terminal_reporter.stats.items():
            if key in ["passed", "failed"]:
                for test_report in value:
                    test_information = PytestTstItem(test_report)
                    test_path = test_information.get_path()
                    test_method_name = test_information.get_method_name()

                    test_name = test_information.get_test_name()
                    exception_file_path = ""
                    exception_line_number = -1
                    target = False
                    if key == "failed":
                        if (
                            legacy_input.get_targeted_failing_test_list_legacy(
                                self._targeted_failing_test_list
                            )
                            is not None
                        ):
                            target = legacy_input.get_targeted_failing_test_list_legacy(
                                self._targeted_failing_test_list
                            ).is_target_test(test_path, test_method_name)
                        elif (
                            legacy_input.get_targeted_failing_test_list_legacy(
                                self._targeted_failing_test_list
                            )
                            is None
                        ):
                            target = True

                        # TODO: Try not to use testReport.longrepr.reprtraceback. Does not work with {pytest-xdist,
                        #  --forked}. Same problem in collect mode and mbfl.
                        repr_traceback = test_report.longrepr.reprtraceback
                        (
                            exception_file_path,
                            exception_line_number,
                        ) = self._traceback_parser.get_exception_location(
                            repr_traceback,
                            self._target_src,
                            self._exclude_list,
                        )
                    self._db_manager.insert_test_case(
                        test_name=test_name,
                        test_type=key,
                        exception_file_path=self._path_util.relative_path_to_abs_path(
                            exception_file_path
                        ),
                        exception_line_number=exception_line_number,
                        target=target,
                    )

        # TODO: Get time from passing tests and target failing tests.
        max_test_time = self._db_manager.select_max_test_time()

        timeout_limit = timeout.get_timeout(max_test_time)
        scored_entity_list = self._ps_run_manager.run_predicate_switching(
            self._target_src,
            self._exclude_list,
            self._fl_granularity,
            timeout_limit,
            self._top_n,
            legacy_input.get_targeted_failing_test_list_legacy(
                self._targeted_failing_test_list
            ),
        )

        self._db_manager.end()
        self._function_level_db_manager.end()

        return scored_entity_list



import os

from fauxpy.command_line.pytest_mode.fl_option_manager import FlOptionManager
from fauxpy.session_lib.fauxpy_session_type import FauxpySessionType
from fauxpy.version import __version__


class PytestOptionManager:
    """
    Manages the command-line options for FauxPy in pytest.
    """

    def __init__(self):
        """
        Initializes the PytestOptionManager with the project working directory.
        """
        self._project_working_directory = os.getcwd()

    @staticmethod
    def add_fauxpy_options(pytest_option_parser):
        """
        Adds FauxPy-specific command-line options to the pytest option parser.

        Args:
            pytest_option_parser: The pytest parser object.
        """
        group = pytest_option_parser.getgroup(
            "fauxpy",
            description="FauxPy " + __version__ + " - An automated fault localization tool for Python programs"
        )
        group.addoption(
            "--src",
            help="Specify the source directory for fault localization. If not provided, FauxPy will be disabled, and pytest will run normally."
        )
        group.addoption(
            "--exclude",
            default="[]",
            help="Provide a list of paths to be excluded from analysis. Default is an empty list."
        )
        group.addoption(
            "--family",
            default="sbfl",
            help="Select the fault localization family to use. Options are: \n"
                 "sbfl (Spectrum-Based Fault Localization), \n"
                 "mbfl (Mutation-Based Fault Localization), \n"
                 "ps (Predicate Switching), \n"
                 "st (Stack-Trace). Default is sbfl."
        )
        group.addoption(
            "--mutation",
            default="t",
            help="Specify the mutation generation strategy for Mutation-Based Fault Localization (MBFL). Options: \n"
                 "t - Use Cosmic Ray with traditional mutation operators (default), \n"
                 "tgpt4ominiapi - Use Cosmic Ray, and when it cannot generate a mutant for a statement, fall back to GPT-4o-mini via its API, \n"
                 "gpt4ominiapi - Use only GPT-4o-mini via its API for mutant generation, without Cosmic Ray, \n"
                 "tgpt4oapi - Use Cosmic Ray, and when it cannot generate a mutant for a statement, fall back to GPT-4o via its API, \n"
                 "gpt4oapi - Use only GPT-4o via its API for mutant generation, without Cosmic Ray."
        )
        group.addoption(
            "--granularity",
            default="statement",
            help="Set the granularity level for fault localization. Options are: \n"
                 "statement (or s) for statement-level analysis, \n"
                 "function (or f) for function-level analysis. Default is statement."
        )
        group.addoption(
            "--top-n",
            default="-1",
            help="Specify the number of top suspicious code elements to report. "
                 "Provide a positive integer to limit the results, or -1 to include all elements. Default is -1."
        )
        group.addoption(
            "--failing-file",
            default=None,
            help="Path to a file containing the list of targeted failing tests. Each line should represent a test identifier. Default is None."
        )
        group.addoption(
            "--failing-list",
            default=None,
            help="Provide a list of targeted failing tests. Default is None."
        )
        group.addoption(
            "--fauxpy-verbose",
            action="store_true",
            default=False,
            help="Show detailed output from FauxPy plugin.",
        )

    def get_fl_option_manager(self, pytest_config) -> FlOptionManager:
        """
        Retrieves a FlOptionManager configured with the options from pytest.

        Args:
            pytest_config: The pytest configuration object.

        Returns:
            FlOptionManager: An instance of FlOptionManager configured with the current pytest options.
        """
        target_src_opt = pytest_config.getoption("--src")
        exclude_list_opt = pytest_config.getoption("--exclude")
        fl_family_opt = pytest_config.getoption("--family")
        mutation_strategy_opt = pytest_config.getoption("--mutation")
        fl_granularity_opt = pytest_config.getoption("--granularity")
        top_n_opt = pytest_config.getoption("--top-n")
        failing_file_opt = pytest_config.getoption("--failing-file")
        failing_list_opt = pytest_config.getoption("--failing-list")
        fauxpy_verbose_opt = pytest_config.getoption("--fauxpy-verbose")
        file_or_dir = pytest_config.getoption("file_or_dir")

        fl_option_manager = FlOptionManager(
            self._project_working_directory,
            target_src_opt,
            exclude_list_opt,
            fl_family_opt,
            mutation_strategy_opt,
            fl_granularity_opt,
            top_n_opt,
            failing_file_opt,
            failing_list_opt,
            fauxpy_verbose_opt,
            file_or_dir,
        )

        return fl_option_manager

    @staticmethod
    def get_fauxpy_session_type(pytest_config):
        """
        Determines the type of FauxPy session based on pytest options.

        Args:
            pytest_config: The pytest configuration object.

        Returns:
            FauxpySessionType: The type of FauxPy session.
        """
        if pytest_config.getoption("--src") is not None:
            return FauxpySessionType.FaultLocalization
        else:
            return FauxpySessionType.FauxpyNotCalled



"""
Module for representing a targeted failing test case in a fault localization session.

A targeted failing test is a failing test that a fault localization technique focuses on
to isolate and locate a specific bug in the presence of multiple bugs.
"""

from typing import Optional

from fauxpy.session_lib.fauxpy_path import FauxpyPath


class TargetedFailingTst:
    """Represents a failing test case targeted for fault localization.

    This class identifies a specific test case
    that triggers a bug, enabling fault localization to isolate it from other failures.
    """

    def __init__(
            self,
            module_path: FauxpyPath,
            class_name: Optional[str],
            function_name: str
    ):
        """Initializes a TargetedFailingTst.

        Args:
            module_path (FauxpyPath): Path to the test module.
            class_name (Optional[str]): Name of the test class, if any.
            function_name (str): Name of the test function.
        """

        self._module_path = module_path
        self._class_name = class_name
        self._function_name = function_name

    def get_relative_test_name(self) -> str:
        """Returns the test's relative name in Pytest-compatible format.

        Returns:
            str: Relative test name in the format
                'module_path::class_name::function_name' or
                'module_path::function_name' if the test is not part of a class.
        """
        if self._class_name is None:
            return f"{self._module_path.get_relative()}::{self._function_name}"
        return f"{self._module_path.get_relative()}::{self._class_name}::{self._function_name}"

    def __str__(self) -> str:
        """Returns the string representation of the targeted test.

        Returns:
            str: Human-readable identifier of the test.
        """
        return self.get_relative_test_name()



import ast
from _ast import FunctionDef
from typing import Any


class FunctionVisitor(ast.NodeVisitor):
    def __init__(self):
        self.function_range_list = []

    def visit_stmt(self, node):
        self.generic_visit(node)

    def visit_FunctionDef(self, node: FunctionDef) -> Any:
        function_name = node.name
        line_start = node.lineno
        line_end = self._get_ending_line(node)
        self.function_range_list.append((function_name, line_start, line_end))
        self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node: FunctionDef) -> Any:
        function_name = node.name
        line_start = node.lineno
        line_end = self._get_ending_line(node)
        self.function_range_list.append((function_name, line_start, line_end))
        self.generic_visit(node)

    # TODO: This method is copied in other places. Combine them somehow.
    # TODO: if node.end_lineno exists, use it. Otherwise, the following one.

    def _get_ending_line(self, node):
        line_end_max = -1

        if hasattr(node, "__dict__"):
            all_att = node.__dict__
            for att in all_att:
                att_instance = node.__getattribute__(att)
                if isinstance(att_instance, list):
                    for sub_att in att_instance:
                        line_end_max = max(self._get_ending_line(sub_att), line_end_max)
                else:
                    line_end_max = max(
                        self._get_ending_line(att_instance), line_end_max
                    )
                if hasattr(att_instance, "lineno"):
                    line_end_max = max(att_instance.lineno, line_end_max)

        if hasattr(node, "lineno"):
            line_end_max = max(node.lineno, line_end_max)
        return line_end_max



from enum import Enum


class FauxpySessionType(Enum):
    """Represents the type of a FauxPy session.

    Attributes:
        FaultLocalization: A session where FauxPy is actively used for fault localization.
        FauxpyNotCalled: A session where FauxPy was not invoked during the test run.
    """
    FaultLocalization = 1
    FauxpyNotCalled = 2



from fauxpy import constants


def get_timeout(max_test_time: float) -> float:
    if max_test_time is None:
        max_test_time = 0
    timeout = constants.TEST_TIMEOUT_SCALING_FACTOR * max_test_time + constants.TEST_TIMEOUT_OFFSET
    return timeout


def get_process_timeout(num_all_tests, timeout_limit):
    return (num_all_tests + 1) * timeout_limit



import pathlib
import shutil
import tempfile
import time


class TempManager:
    def __init__(self, project_working_directory: pathlib.Path):
        self._project_working_directory = project_working_directory

    @staticmethod
    def _get_unique_temporary_directory_path() -> str:
        tmp = tempfile.gettempdir()
        current_time_int = time.time()
        tmp_dir = pathlib.Path(tmp) / str(current_time_int)
        tmp_dir = str(tmp_dir.resolve())
        return tmp_dir

    @staticmethod
    def _copy_dir(source: str, destination: str):
        shutil.copytree(source, destination, symlinks=True)

    def make_project_copy_in_temp(self) -> str:
        temp_dir = self._get_unique_temporary_directory_path()
        project_dir = str(self._project_working_directory.absolute().resolve())
        self._copy_dir(project_dir, temp_dir)

        return temp_dir



from pathlib import Path

from fauxpy.fault_localization.collect_ps_info.db_manager import (
    CollectPsInfoDbManager,
)
from fauxpy.fault_localization.collect_ps_info.reporter_lib import CollectPsInfoReporter
from fauxpy.fault_localization.collect_ps_info.session_file import (
    CollectPsInfoSessionFileManager,
)
from fauxpy.session_lib.fl_session import FlSession
from fauxpy.session_lib.pytest_tst_item import PytestTstItem


class CollectPsInfoSession(FlSession):
    def __init__(self, report_directory_path: Path, project_working_directory: Path):
        self._current_test_name = None
        self._db_manager = CollectPsInfoDbManager(report_directory_path)
        self._reporter = CollectPsInfoReporter(self._db_manager)
        self._session_file_manager = CollectPsInfoSessionFileManager(
            report_directory_path, project_working_directory
        )

    @staticmethod
    def __pretty_representation():
        return "Collect PS Info Mode"

    def __str__(self):
        return self.__pretty_representation()

    def __repr__(self):
        return self.__pretty_representation()

    def run_test_call(self, item):
        self._current_test_name = PytestTstItem(item).get_test_name()

    def run_test_make_report(self, item, call):
        if call.when == "call":
            test_name = PytestTstItem(item).get_test_name()
            if test_name != self._current_test_name:
                raise Exception(
                    f"Starting with {self._current_test_name}, but ending with {test_name}."
                )

            predicate_sequence = (
                self._session_file_manager.load_executed_predicate_sequence_and_remove_file()
            )
            if predicate_sequence is not None:
                self._db_manager.insert_predicate_sequence(
                    test_name, predicate_sequence
                )

    def terminal_summary(self, terminal_reporter, exit_status):
        predicate_sequence_table = (
            self._reporter.get_json_test_predicate_sequence_table()
        )
        self._session_file_manager.save_predicate_sequence_table(
            predicate_sequence_table
        )
        self._db_manager.end()
        return {"Default": []}



import sqlite3
from pathlib import Path


class CollectMbflDBManager:
    _File_name = "fauxpy.db"
    _Test_case_table = "TestCase"

    def __init__(self, report_directory_path: Path):
        db_file_path = report_directory_path / self._File_name
        db_file_path_str = str(db_file_path.absolute().resolve())
        self._connection = sqlite3.connect(db_file_path_str)
        command_list = self._get_database_schema()
        for command in command_list:
            self._connection.executescript(command)

    def _get_database_schema(self):
        # TODO: Add indices.
        test_case_table_create_command = (
            f"CREATE TABLE {self._Test_case_table} "
            f"(Rowid INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, "
            f"TestName TEXT NOT NULL UNIQUE, "
            f"Type TEXT NOT NULL, "
            f"ShortTraceback TEXT NOT NULL,"
            f"Timeout INTEGER NOT NULL);"
        )

        command_list = [
            test_case_table_create_command,
        ]

        return command_list

    def end(self):
        self._connection.close()

    def insert_test_case(
        self, test_name: str, test_type: str, short_traceback: str, timeout_stat: int
    ):
        cur = self._connection.cursor()
        cur.execute(
            f"INSERT INTO {self._Test_case_table} VALUES (NULL, ?, ?, ?, ?)",
            (test_name, test_type, short_traceback, timeout_stat),
        )

        self._connection.commit()

    def select_all_test_cases(self):
        cur = self._connection.cursor()
        cur.execute(
            f"SELECT TestName, Type, ShortTraceBack, Timeout FROM {self._Test_case_table}"
        )
        rows = cur.fetchall()

        return rows



import csv
import json
from pathlib import Path
from typing import List, Tuple

from fauxpy import constants
from fauxpy.session_lib.fauxpy_path import FauxpyPath
from fauxpy.session_lib.fl_type import FlFamily, FlGranularity, MutationStrategy
from fauxpy.session_lib.targeted_failing_tst import TargetedFailingTst


class FlFileManager:
    """Manages file operations of a fault localization session."""
    def __init__(self, report_directory_path: Path):
        """
        Initializes the file manager with a directory path for storing reports.

        Args:
            report_directory_path (Path): Path to the directory where fault localization reports will be saved.
        """
        self._report_directory_path = report_directory_path

    def save_scores_to_file(self, technique: str, scores: List[Tuple[str, int]]):
        """
        Saves fault localization scores to a CSV file.

        Args:
            technique (str): Name of the fault localization technique used.
            scores (List[Tuple[str, int]]): List of tuples containing (entity name, score).
        """
        scores_file_path = self._report_directory_path / f"Scores_{technique}.csv"
        with open(scores_file_path, "w") as csvFile:
            writer = csv.writer(csvFile)
            writer.writerow(constants.SCORES_CSV_HEADER)
            for score in scores:
                writer.writerow([score[0], score[1]])

    def save_config_to_file(
        self,
        target_src: FauxpyPath,
        exclude_list: List[FauxpyPath],
        fl_family: FlFamily,
        fl_granularity: FlGranularity,
        top_n: int,
        targeted_failing_test_list: List[TargetedFailingTst],
        mutation_strategy: MutationStrategy,
    ):
        """
        Saves the session configuration to a JSON file.

    Args:
        target_src (FauxpyPath): Source code directory to perform fault localization on.
        exclude_list (List[FauxpyPath]): List of files or directories to exclude from fault localization.
        fl_family (FlFamily): Family of the fault localization technique.
        fl_granularity (FlGranularity): Fault localization granularity level.
        top_n (int): Number of top-ranked elements to show in the output.
        targeted_failing_test_list (List[TargetedFailingTst]): List of failing test cases that are the target
            of the current fault localization session.
        mutation_strategy (MutationStrategy): Strategy used for mutation analysis
            (only used by the MBFL family).
        """
        file_path = self._report_directory_path / constants.CONFIG_FILE_NAME_FL_SESSION

        config_dict = {
            "Src": target_src.get_relative(),
            "Exclude": [x.get_relative() for x in exclude_list],
            "Family":  fl_family.name,
            "Granularity": fl_granularity.name,
            "TopN": top_n,
            "TargetedFailingTests": [x.get_relative_test_name() for x in targeted_failing_test_list],
            "MutationStrategy": mutation_strategy.name
        }

        with file_path.open("w", encoding="utf-8") as file:
            json.dump(config_dict, file, indent=4)  # Pretty-print with indentation

    def save_delta_time_to_file(self, delta_time):
        """
        Saves the fault localization session execution time (delta time) to a JSON file.

        Args:
            delta_time (float): Time taken by the fault localization session.
        """
        file_path = self._report_directory_path / constants.TIME_FILE_NAME_FL_SESSION
        delta_time_dict = {
            "DeltaTime": delta_time
        }
        with file_path.open("w", encoding="utf-8") as file:
            json.dump(delta_time_dict, file, indent=4)  # Pretty-print with indentation

    def get_log_file_path(self):
        """
        Returns the absolute path of the log file.

        Returns:
            str: Absolute path to the log file.
        """
        file_path = self._report_directory_path / constants.LOG_FILE_NAME_FL_SESSION
        return str(file_path.absolute().resolve())



import json

from fauxpy.fault_localization.collect_ps_info.db_manager import CollectPsInfoDbManager


class CollectPsInfoReporter:
    def __init__(self, db_manager: CollectPsInfoDbManager):
        self._db_manager = db_manager

    # def save_test_predicate_sequence_table(self):
    #     table = self._db_manager.select_all_predicate_sequences()
    #     json_table = json.dumps(table)
    #     common.saveInCollectModePredicateSequenceTable(json_table)

    def get_json_test_predicate_sequence_table(self):
        table = self._db_manager.select_all_predicate_sequences()
        json_table = json.dumps(table)
        return json_table



import datetime
import shutil
from pathlib import Path

from fauxpy import constants
from .fauxpy_path import FauxpyPath
from .fl_type import FlFamily, FlGranularity, MutationStrategy


class FlPathManager:
    """A class that manages paths for a fault localization session."""

    def __init__(
        self,
        current_working_directory: FauxpyPath,
        fl_family: FlFamily,
        fl_granularity: FlGranularity,
        mutation_strategy: MutationStrategy
    ):
        """
        Initializes the class instance.

        Args:
            current_working_directory (FauxpyPath): The directory of the project that is the subject of fault localization.
            fl_family (FlFamily): The fault localization family.
            fl_granularity (FlGranularity): The granularity of fault localization.
            mutation_strategy (MutationStrategy): The mutation strategy used for generating mutations (only used by MBFL).
        """
        self._current_working_directory = Path(current_working_directory.get_absolute())
        self._fl_family = fl_family
        self._fl_granularity = fl_granularity
        self._mutation_strategy = mutation_strategy
        self._report_directory_path = None

    def get_report_directory_path(self) -> Path:
        """
        Returns the path to the fault localization report directory.
        If the report directory does not exist, it is created.

        Returns:
            Path: The path to the generated report directory.
        """
        if self._report_directory_path is None:
            self._report_directory_path = self._create_report_directory()

        return self._report_directory_path

    def _create_report_directory(self):
        """
        Creates and returns a new report directory based on the current working directory and other attributes.

        Returns:
            Path: The path to the newly created report directory.
        """
        current_working_dir_name = self._current_working_directory.name
        current_working_dir_parent_path = self._current_working_directory.parent
        ft = "%Y_%m_%d_%H_%M_%S_%f"
        current_data_time = datetime.datetime.now().strftime(ft)
        report_director_name = (
            f"{constants.REPORT_DIRECTORY_NAME_PREFIX_FL_SESSION}_"
            f"{current_working_dir_name}_"
            f"{self._fl_family.name.lower()}_"
            f"{self._fl_granularity.name.lower()}_"
            f"{self._mutation_strategy.name.lower()}_"
            f"{current_data_time}"
        )
        report_directory_path = current_working_dir_parent_path / report_director_name

        if self._fl_family in [
            FlFamily.CollectMbfl,
            FlFamily.CollectPsInfo,
            FlFamily.CollectPsRun,
        ]:
            report_directory_path = (
                self._current_working_directory
                / constants.FileNames.CollectModeDirectoryName
            )

        if report_directory_path.exists():
            shutil.rmtree(report_directory_path)  # Remove last reports

        report_directory_path.mkdir()

        return report_directory_path



from typing import Tuple, List


class Mutant:
    """Represents a mutant in a source code file, created by applying a mutation operator."""

    def __init__(
        self,
        module_path: str,
        operator_name: str,
        start_pos: Tuple[int, int],
        end_pos: Tuple[int, int],
        occurrence: int,
        module_content: str,
        module_diff: List[str],
    ):
        """
        Initializes a Mutant instance.

        Args:
            module_path (str): The path to the source code module.
            operator_name (str): The name of the mutation operator applied.
            start_pos (Tuple[int, int]): The starting position (line, column) of the mutation.
            end_pos (Tuple[int, int]): The ending position (line, column) of the mutation.
            occurrence (int): The occurrence count of the mutation in the module.
            module_content (str): The content of the module after mutation.
            module_diff (List[str]): The mutation as diff.
        """
        self._module_path: str = module_path
        self.operator_name: str = operator_name
        self.start_pos: Tuple[int, int] = start_pos
        self.end_pos: Tuple[int, int] = end_pos
        self._occurrence: int = occurrence
        self._module_content: str = module_content
        self._module_diff: List[str] = module_diff
        self.id = "-"

    def set_id(self, val):
        """
        Sets the unique identifier for the mutant.

        Args:
            val (str): The identifier to be assigned.
        """
        self.id = val

    def get_id(self):
        """Returns the unique identifier of the mutant."""
        return self.id

    def get_module_path(self):
        """Returns the path of the module where the mutation occurred."""
        return self._module_path

    def get_module_operator(self):
        """Returns the name of the mutation operator applied."""
        return self.operator_name

    def get_module_content(self):
        """Returns the content of the module after mutation."""
        return self._module_content

    def get_module_diff_as_text(self):
        """Returns the mutation as diff."""
        diff_text = "\r\n".join(self._module_diff)
        return diff_text

    def operator_mutation_target_equal_to(self, mutant):
        """
        Checks if the mutation targets the same location and operator as another mutant.

        Args:
            mutant (Mutant): Another mutant to compare against.

        Returns:
            bool: True if both mutants target the same location and operator, False otherwise.
        """
        result = (
            self.operator_name == mutant.operator_name
            and self.start_pos == mutant.start_pos
            and self.end_pos == mutant.end_pos
        )
        return result

    # TODO: Multiline predicates might cause problem.
    def get_line_number(self):
        # A predicate can spread through multiple lines. In such cases,
        # we consider each line of the predicate a separate statement
        # while performing fault localization.
        # For instance, in the following predicate, we have two different lines
        # to mutate:
        # 12: if x > 10 and \
        # 13:   y < 24:
        # 14   ...
        # A mutation applied on line 12 or 13 (by for instance
        # changing the relational operators) are considered two
        # mutants with LineNumber == self.startPos[0].
        # However, putting a not operator after if affects the whole
        # predicate (i.e., both lines 12 and 13). In this case,
        # self.startPos[0] != self.endPos[0].
        # But even in this case, we consider LineNumber == self.startPos[0].
        # Thus, LineNumber is always self.startPos[0] (it is our assumption).

        return self.start_pos[0]

    def get_start_pos(self):
        """Returns the starting line number of the mutation."""
        return self.start_pos[0]

    def get_end_pos(self):
        """Returns the ending line number of the mutation."""
        return self.end_pos[0]



from typing import List, Tuple, Optional


class CollectPsRunResult:
    def __init__(
        self,
        test_case_table: List[Tuple[str, str, str]],
        seen_exception_list: List[Tuple[str, str]],
    ):
        self.testCaseTable = test_case_table
        self.seenExceptionList = seen_exception_list

    def get_test_result(self, test_name: str) -> Tuple[Optional[str], Optional[float]]:
        for currentTest, testType, stacktrace, timeoutStat in self.testCaseTable:
            if currentTest == test_name:
                return testType, timeoutStat
        return None, None

    def get_test_seen_exception_list(self, test_name: str) -> str:
        for item in self.seenExceptionList:
            if item[0] == test_name:
                return item[1]
        return ""

    def is_test_case_table_empty_or_none(self):
        return self.testCaseTable is None or len(self.testCaseTable) == 0



from fauxpy.fault_localization.mbfl.metric_metallaxis import MetricMetallaxis
from fauxpy.fault_localization.mbfl.db_manager import MbflDbManager
from fauxpy.fault_localization.mbfl.metric_muse import MetricMuse


class MutantScoreManager:
    EPSILON = 0.01

    def __init__(self, db_manager: MbflDbManager):
        self._db_manager = db_manager
        self._metric_muse = MetricMuse(self.EPSILON)
        self._metric_metallaxis = MetricMetallaxis()

    def compute_mutant_scores_store_db(self):
        (
            total_failed_to_passed,
            total_passed_to_failed,
        ) = self._db_manager.select_total_failed_to_passed_and_passed_to_failed()

        _, num_all_failed = self._db_manager.select_number_of_tests()

        mutants = self._db_manager.select_mutant_score_terms()
        for mutant in mutants:
            (
                mutant_id,
                mutant_failed_to_pass,
                mutant_passed_to_failed,
                mutant_failed_changed,
            ) = mutant

            muse_score = self._metric_muse.compute_mutant_score(
                mutant_failed_to_pass,
                mutant_passed_to_failed,
                total_failed_to_passed,
                total_passed_to_failed,
            )

            metallaxis_score = self._metric_metallaxis.compute_mutant_score(
                mutant_failed_to_pass,
                mutant_failed_changed,
                mutant_passed_to_failed,
                num_all_failed,
            )

            self._db_manager.update_mutant_score_terms(
                mutant_id, muse_score, metallaxis_score
            )



class RankManager:
    def __init__(self, sb_manager):
        self._sb_manager = sb_manager

    @staticmethod
    def compute_score_list(traceback_function_name_list):
        scores = []
        for j in range(len(traceback_function_name_list)):
            item = (traceback_function_name_list[j], float(1) / (j + 1))
            scores.append(item)

        return scores

    def get_sorted_score_list(self, top_n: int):
        if top_n == -1:
            scored_function_list = self._sb_manager.select_all_ranked_functions()
        elif top_n >= 1:
            scored_function_list = self._sb_manager.select_top_n_ranked_functions(top_n)
        else:
            raise Exception(f"TopN {top_n} is not supported.")

        return scored_function_list



import pytest
from unittest.mock import MagicMock

from fauxpy.session_lib.targeted_failing_tst import TargetedFailingTst


@pytest.fixture
def mock_fauxpy_path():
    """Creates a mock FauxpyPath that returns a relative module path."""
    mock = MagicMock()
    mock.get_relative.return_value = "tests/test_example.py"
    return mock


def test_relative_name_with_class(mock_fauxpy_path):
    """Should include class name in the relative test identifier."""
    test_case = TargetedFailingTst(
        module_path=mock_fauxpy_path,
        class_name="TestExample",
        function_name="test_function"
    )
    expected = "tests/test_example.py::TestExample::test_function"
    assert test_case.get_relative_test_name() == expected
    assert str(test_case) == expected


def test_relative_name_without_class(mock_fauxpy_path):
    """Should omit class name when it is None."""
    test_case = TargetedFailingTst(
        module_path=mock_fauxpy_path,
        class_name=None,
        function_name="test_function"
    )
    expected = "tests/test_example.py::test_function"
    assert test_case.get_relative_test_name() == expected
    assert str(test_case) == expected



from pathlib import Path

import setuptools
from setuptools import setup


# Load version
root_path = Path(__file__).parent
version_file_path = root_path / "fauxpy" / "version.py"
__version__ = ""
exec(version_file_path.read_text())
assert __version__ != ""

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="fauxpy",
    version=__version__,
    author="Mohammad Rezaalipour",
    author_email="rezaalipour.mohammad@gmail.com",
    description="An automated fault localization tool for Python programs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://fauxpy.readthedocs.io",
    packages=setuptools.find_packages(exclude="tests"),
    python_requires=">=3.8",
    install_requires=[
        "pytest>=3.1.2",
        "coverage>=6.2",
        "cosmic-ray==8.3.5",
        "astor~=0.8.1",
        "pytest-timeout==2.1.0",
        "wheel",
        "pyllmut"
    ],
    entry_points={
        "pytest11": [
            "fauxpy = fauxpy.main",
        ],
        "console_scripts": ["fauxpy = fauxpy.main:fauxpy_analysis_mode"],
    },
)



import json
import pathlib
from typing import Optional, List, Tuple

from fauxpy import constants
from fauxpy.constants import FileNames


class CollectPsRunApiFileManager:
    @staticmethod
    def load_test_case_table(
        project_path: str,
    ) -> Optional[List[Tuple[str, str, str]]]:
        file_path = (
            pathlib.Path(project_path)
            / FileNames.CollectModeDirectoryName
            / FileNames.collectModeTestCases
        )
        if not file_path.exists():
            return None

        with open(file_path, "r") as file:
            json_table = file.read()
            table = json.loads(json_table)
        return table

    @staticmethod
    def load_seen_exception_sequence_table(
        project_path,
    ) -> Optional[List[Tuple[str, str]]]:
        file_path = (
            pathlib.Path(project_path)
            / FileNames.CollectModeDirectoryName
            / FileNames.collectModeSeenExceptions
        )
        if not file_path.exists():
            return None

        with open(file_path, "r") as file:
            json_table = file.read()
            table = json.loads(json_table)
        return table

    @staticmethod
    def save_config_file(project_path: str, predicate_name: str, instance_number: int):
        config_file_name = constants.getCollectModeConfigFileName()
        config_file_path = pathlib.Path(project_path) / pathlib.Path(config_file_name)
        content_dictionary = {
            "PredicateName": predicate_name,
            "InstanceNumber": instance_number,
        }
        json_content = json.dumps(content_dictionary)
        with open(config_file_path, "w") as file:
            file.write(json_content)



from typing import List, Tuple

from fauxpy.fault_localization.ps.db_manager import PsDbManager


class SeenExceptionManager:
    def __init__(self, db_manager: PsDbManager):
        self._db_manager = db_manager

    def _get_seen_exceptions_store_db(self, exceptions: List[Tuple[str, str, int]]):
        for i, exp in enumerate(exceptions):
            test_name, file_path, line_number = exp
            exception_name = f"Exception_{i}"

            self._db_manager.insert_seen_exceptions(
                test_name=test_name,
                file_path=file_path,
                line_number=line_number,
                exception_name=exception_name,
            )

    def get_seen_exceptions_store_db(self):
        """
        Stores in database the info about exception raises
        that happen while running the failing tests.
        """

        exceptions = self._db_manager.select_test_case_exceptions()
        self._get_seen_exceptions_store_db(exceptions)



import time

class Timer(object):
    """A simple timer class to measure elapsed time."""

    def __init__(self):
        """Initializes the Timer with a starting time of zero."""
        self._starting_time = 0

    def start(self):
        """Starts the timer by recording the current time."""
        self._starting_time = time.time()

    def end(self) -> float:
        """Ends the timer and returns the elapsed time.

        Returns:
            float: The number of seconds elapsed since the timer was started.
        """
        ending_time = time.time()
        return ending_time - self._starting_time



import sqlite3
from pathlib import Path
from typing import List, Tuple


class FunctionLevelDbManager:
    _File_name = "common.db"
    _Function_info_table = "FunctionInformation"

    def __init__(self, report_directory_path: Path):
        db_file_path = report_directory_path / self._File_name
        db_file_path_str = str(db_file_path.absolute().resolve())
        self._connection = sqlite3.connect(db_file_path_str)
        commands = self._get_database_schema()
        for command in commands:
            self._connection.executescript(command)

    def _get_database_schema(self):
        function_info_table_create_command = (
            f"CREATE TABLE {self._Function_info_table} "
            f"(Rowid INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, "
            f"FilePath	TEXT NOT NULL, "
            f"FunctionName	TEXT NOT NULL, "
            f"LineStart INTEGER NOT NULL, "
            f"LineEnd INTEGER NOT NULL);"
        )

        function_info_table_index_command = f"CREATE INDEX Index_OnThree ON {self._Function_info_table} (FilePath, LineStart, LineEnd);"

        commands = [
            function_info_table_create_command,
            function_info_table_index_command,
        ]

        return commands

    def end(self):
        self._connection.close()

    def select_function_ranges(
        self, file_path: str, line_number: int
    ) -> List[Tuple[str, int, int]]:
        cur = self._connection.cursor()

        cur.execute(
            f"SELECT FunctionName, LineStart, LineEnd FROM {self._Function_info_table} "
            f"WHERE FilePath = ? and LineStart <= ? and LineEnd >= ?",
            (file_path, line_number, line_number),
        )

        var = cur.fetchall()

        return var

    def insert_function_information(
        self, file_path: str, function_name: str, line_start: int, line_end: int
    ):
        cur = self._connection.cursor()

        cur.execute(
            f"INSERT INTO {self._Function_info_table} VALUES (NULL, ?, ?, ?, ?)",
            (file_path, function_name, line_start, line_end),
        )
        self._connection.commit()

    def is_file_path_covered(self, file_path: str) -> bool:
        cur = self._connection.cursor()

        cur.execute(
            f"SELECT COUNT(*) FROM {self._Function_info_table} WHERE FilePath = ?",
            (file_path,),
        )

        cnt = cur.fetchone()

        return cnt[0] > 0



import shutil
from pathlib import Path
from typing import List, Tuple, Optional

from fauxpy.fault_localization.collect_ps_info.api_lib import CollectPsInfoApi
from fauxpy.fault_localization.collect_ps_run.api_lib import CollectPsRunApi
from fauxpy.fault_localization.granularity.db_manager import FunctionLevelDbManager
from fauxpy.fault_localization.granularity.function_level import (
    FunctionLevelGranularity,
)
from fauxpy.fault_localization.ps.ast_lib.instrumentation_manager import (
    InstrumentationManager,
)
from fauxpy.fault_localization.ps.db_manager import PsDbManager
from fauxpy.fault_localization.ps.predicate_instance_manager import (
    PredicateInstanceManager,
)
from fauxpy.fault_localization.ps.seen_exception_manager import SeenExceptionManager
from fauxpy.fault_localization.util.path_util import PathUtil
from fauxpy.fault_localization.util.temp_lib import TempManager
from fauxpy.session_lib import naming_lib
from fauxpy.session_lib.fauxpy_path import FauxpyPath
from fauxpy.session_lib.fauxpy_printer import fl_print
from fauxpy.session_lib.fl_type import FlGranularity
from fauxpy.session_lib.target_tsts import TargetFailingTests


class PsRunManager:
    def __init__(
        self,
        db_manager: PsDbManager,
        function_level_db_manager: FunctionLevelDbManager,
        project_working_directory: Path,
    ):
        self._db_manager = db_manager
        self._seen_exception_manager = SeenExceptionManager(self._db_manager)
        self._predicate_instance_manager = PredicateInstanceManager(self._db_manager)
        self._instrumentation_manager = InstrumentationManager()
        self._collect_ps_info_api = CollectPsInfoApi()
        self._collect_ps_run_api = CollectPsRunApi()
        self._function_level_granularity_manager = FunctionLevelGranularity(
            function_level_db_manager
        )
        self._temp_manager = TempManager(project_working_directory)
        self._path_util = PathUtil(project_working_directory)

    def _get_instrumented_module_list(self) -> List[Tuple[str, str]]:
        """
        Returns a list of tuples of modules paths (having at least one candidate
        predicate or a seen exception) and their instrumented content.
        """
        inst_file_paths_content: List[Tuple[str, str]] = []
        file_paths_with_candidate_predicate = (
            self._db_manager.select_distinct_candidate_predicate_file_paths()
        )
        file_paths_with_seen_exceptions = (
            self._db_manager.select_distinct_seen_exceptions_file_paths()
        )
        inst_file_paths = set(
            file_paths_with_candidate_predicate + file_paths_with_seen_exceptions
        )
        for filePath in inst_file_paths:
            candidate_predicates = (
                self._db_manager.select_candidate_predicates_for_file_path(filePath)
            )
            seen_exception_list = self._db_manager.select_seen_exceptions_for_file_path(
                filePath
            )
            inst_file_content = (
                self._instrumentation_manager.instrument_current_file_path(
                    filePath, candidate_predicates, seen_exception_list
                )
            )
            if inst_file_content is None:
                # For cases that astor cannot perform correctly
                for item in candidate_predicates:
                    line_start, line_end, predicate_name = item
                    self._db_manager.insert_astor_assert_error_info(
                        filePath, line_start, line_end, predicate_name
                    )
            else:
                inst_file_paths_content.append((filePath, inst_file_content))

        return inst_file_paths_content

    def _inject_instrumented_content_in_project(
        self, temp_project_path: str, instrumented_modules: List[Tuple[str, str]]
    ):
        """
        Replaces modules to be instrumented with their instrumented versions.
        """
        for modulePath, moduleContent in instrumented_modules:
            relative_module_path = self._path_util.absolute_path_to_relative_path(
                modulePath
            )
            absolute_temp_module_path = Path(temp_project_path) / relative_module_path
            with open(absolute_temp_module_path, "w") as file:
                file.write(moduleContent)

    @staticmethod
    def _remove_temp_project(temp_project_path):
        shutil.rmtree(temp_project_path)

    def _instrument_project(self) -> str:
        """
        Makes a copy of the project in a temporary directory, instruments it, and
         returns the path of the project copy.
        """
        inst_file_paths_content = self._get_instrumented_module_list()
        tmp_project_path = self._temp_manager.make_project_copy_in_temp()
        self._inject_instrumented_content_in_project(
            tmp_project_path, inst_file_paths_content
        )
        return tmp_project_path

    def _get_predicate_sequences(
            self,
            project_path: str,
            target_src: FauxpyPath,
            exclude_list: List[FauxpyPath],
            failed_test_paths
    ):
        """
        Runs the project in Collect mode and stores a list of tuples
        (test_name, Predicate sequence).
        @param project_path: path to the project subject to fault localization.
        """

        # TODO: Optimize the failing tests to run. Maybe running it on arg_min(failedTestPaths, fileOrDir).

        predicate_sequences = []
        indexed_predicate_sequences = (
            self._collect_ps_info_api.run_ps_collect_mode_info(
                target_src=target_src,
                exclude_list=exclude_list,
                project_path=project_path,
                file_or_dir=failed_test_paths,
            )
        )
        # exeResultData = collect_mode.getRunResult(projectPath)
        for r in indexed_predicate_sequences:
            test_name, predicate_sequence, ind_pred_seq = r
            predicate_sequences.append((test_name, ind_pred_seq))

            # Not needed.
            self._db_manager.insert_predicate_sequence_for_test(
                test_name, predicate_sequence, ind_pred_seq
            )

        return predicate_sequences

    def _get_generalized_failed_test_function_paths(self):
        """
        Returns failed test paths (i.e., TEST_FILE_PATH::TEST_FUNCTION_NAME).
        For parametrized tests, returns generalized path
         (i.e., parameters excluded).
        """
        generalized_test_names = []
        failed_test_case_names = self._db_manager.select_test_case_failed()
        for testName in failed_test_case_names:
            # generalized_test_name = _getGeneralizedTestName(testName)
            file_path, _, function_name = naming_lib.convert_test_name_to_components(
                testName
            )
            generalized_test_name = naming_lib.get_generalized_test_name(
                file_path, function_name
            )
            generalized_test_names.append(generalized_test_name)
        return generalized_test_names

    def _run_switched_predicate_instance(
        self,
        project_path: str,
        test_name: str,
        predicate_name: str,
        instance_number: int,
        target_src: FauxpyPath,
        exclude_list: List[FauxpyPath],
        timeout_limit: float,
    ) -> Tuple[Optional[str], Optional[List[str]], Optional[float], bool]:
        """
        Runs the instrumented project in temp directory on the given test name.
        In the execution, the given predicate instance is switched.
        Returns True if the given test passes, and False, if it does not.
        """

        # generalized_test_path = _getGeneralizedTestName(testName)
        file_path, _, function_name = naming_lib.convert_test_name_to_components(
            test_name
        )
        generalized_test_path = naming_lib.get_generalized_test_name(
            file_path, function_name
        )
        exe_result_data = self._collect_ps_run_api.run_ps_collect_mode_run(
            target_src=target_src,
            exclude_list=exclude_list,
            project_path=project_path,
            file_or_dir=[generalized_test_path],
            predicate_name=predicate_name,
            instance_number=instance_number,
            timeout=timeout_limit,
        )

        exec_stat_error = exe_result_data.is_test_case_table_empty_or_none()
        if exec_stat_error:
            fl_print.normal("Bad execution")
            return None, None, None, exec_stat_error

        test_result, timeout_stat = exe_result_data.get_test_result(test_name)
        if test_result is None and timeout_stat is None:
            # The parametrized tests might be executed with different parameters in the
            # main mode and the collect mode. In this situation, the test name from
            # main cannot be found in the tests executed in collect mode. Found by pandas 141.
            fl_print.normal("Non-deterministic execution")
            return None, None, None, False

        seen_exception_list_str = exe_result_data.get_test_seen_exception_list(
            test_name
        )
        seen_exception_list = self.csv_row_to_list(seen_exception_list_str)

        return test_result, seen_exception_list, timeout_stat, exec_stat_error

    @staticmethod
    def csv_row_to_list(csv_row: str) -> List[str]:
        csv_comps = csv_row.split(",")
        return csv_comps

    def _run_predicate_sequence(
        self,
        project_path: str,
        test_name: str,
        indexed_pred_seq_str: str,
        target_src: FauxpyPath,
        exclude_list: List[FauxpyPath],
        expected_exception_seen_name: str,
        timeout_limit: float,
    ):
        """
        Runs the given ordered predicate instances for the given predicate test name.
        Returns an ordered sequence of predicate instances that can pass the given test if switched.
        """

        ind_pred_seq = self.csv_row_to_list(indexed_pred_seq_str)
        ind_pred_seq.reverse()  # Predicate instances are executed from last to first

        number_of_predicate_instances_for_test = len(ind_pred_seq)

        passing_predicate_instances = []
        for ind, predInst in enumerate(ind_pred_seq):
            pred_name, inst_num = predInst.split("::")

            fl_print.normal(
                f"Running Predicate Instance "
                f"'{pred_name}::{inst_num}' ({ind + 1}/{number_of_predicate_instances_for_test}) "
                f"with test '{test_name}'"
            )

            (
                test_result,
                seen_exception_list,
                timeout_stat,
                exec_stat_error,
            ) = self._run_switched_predicate_instance(
                project_path=project_path,
                test_name=test_name,
                predicate_name=pred_name,
                instance_number=int(inst_num),
                target_src=target_src,
                exclude_list=exclude_list,
                timeout_limit=timeout_limit,
            )

            if exec_stat_error:
                self._db_manager.insert_bad_execution_predicate_instance(
                    test_name, pred_name, inst_num
                )
                fl_print.normal("Execution error for: ", test_name, pred_name, inst_num)
                continue

            # Not needed. Just to collect info for further analysis.
            if timeout_stat == 1:
                self._db_manager.insert_timeout_predicate_instance(
                    test_name, pred_name, inst_num
                )
                fl_print.normal("Timeout for: ", test_name, pred_name, inst_num)
                continue

            if test_result == "passed":
                # For crashing bugs, if switching prevents the program from crashing
                # but the crash location is not executed, the predicate is not critical.
                if (expected_exception_seen_name is None) or (
                    expected_exception_seen_name in seen_exception_list
                ):
                    passing_predicate_instances.append(predInst)

        return passing_predicate_instances

    def _get_test_scored_entity_store_db(
        self,
        test_name: str,
        passing_predicate_instance_sequence: List[str],
        fl_granularity: FlGranularity,
    ):
        score = 1
        for predicateInstance in passing_predicate_instance_sequence:
            predicate_name, instance_number = predicateInstance.split("::")
            file_path, line_start, line_end = (
                self._db_manager.select_candidate_predicate(predicate_name)
            )

            if fl_granularity == FlGranularity.Statement:
                entity = f"{file_path}::{line_start}::{line_end}"
            elif fl_granularity == FlGranularity.Function:
                cfi = self._function_level_granularity_manager.get_covered_function(
                    file_path, line_start
                )
                if cfi is None:
                    entity = f"{file_path}::GLOBAL"
                else:
                    (
                        function_file_path,
                        function_name,
                        function_line_start,
                        function_line_end,
                    ) = cfi
                    entity = naming_lib.get_covered_function_name(
                        function_file_path,
                        function_name,
                        function_line_start,
                        function_line_end,
                    )
            else:
                # TODO: Safe to remove (handled by input validation).
                raise Exception(f"The granularity {fl_granularity.name} is not supported.")

            # For function granularity it can happen.
            # In this case, only one should be stored.
            # Found in youtube_dl13
            if self._db_manager.scored_entity_exists_for_test(test_name, entity):
                continue

            self._db_manager.insert_scored_entity(test_name, entity, score)

            # Critical predicate closer to the failure location is more suspicious to be buggy.
            # Predicate instance closer to the failure locations is more suspicious to be buggy.
            score = score / 1.1

    def _get_all_tests_top_n_scored_entities(self, top_n):
        all_tests_top_n_scored_entities = {}

        final_failing_test_names = self._db_manager.select_scored_entity_test_names()

        for testName in final_failing_test_names:
            test_name_as_file_path = naming_lib.test_name_to_file_name(testName)

            if top_n == -1:
                test_top_n_scored_entities = (
                    self._db_manager.select_test_all_scored_entities(testName)
                )
            elif top_n >= 0:
                test_top_n_scored_entities = (
                    self._db_manager.select_test_top_n_scored_entities(testName, top_n)
                )
            else:
                raise Exception(f"TopN {top_n} is not supported.")

            all_tests_top_n_scored_entities[test_name_as_file_path] = (
                test_top_n_scored_entities
            )

        if len(all_tests_top_n_scored_entities) == 0:
            fl_print.normal("There were no swapped instances.")
            # all_tests_top_n_scored_entities = {"fauxpy_no_swapped_instances": []}
            all_tests_top_n_scored_entities = {"PS": []}

        return all_tests_top_n_scored_entities

    def run_predicate_switching(
        self,
        target_src: FauxpyPath,
        exclude_list: List[FauxpyPath],
        fl_granularity: FlGranularity,
        timeout_limit: float,
        top_n: int,
        target_failing_tests: TargetFailingTests,
    ):
        """
        Runs the whole predicate switching algorithm.
        """

        self._predicate_instance_manager.get_candidate_predicates_store_db()

        if self._db_manager.number_of_candidate_predicates() == 0:
            fl_print.normal("There were no candidate predicates.")
            # return {"fauxpy_no_candidate_predicates": []}
            return {"PS": []}

        self._seen_exception_manager.get_seen_exceptions_store_db()
        temp_project_path = self._instrument_project()

        if target_failing_tests is None:
            failing_tests = self._get_generalized_failed_test_function_paths()
        else:
            failing_tests = target_failing_tests.get_failing_tests()

        predicate_sequences = self._get_predicate_sequences(
            temp_project_path,
            target_src,
            exclude_list,
            failing_tests
        )

        for pred_seq_item in predicate_sequences:
            test_name, indexed_pred_seq_str = pred_seq_item
            original_test_type = self._db_manager.select_test_type(test_name)
            if original_test_type == "passed":
                continue

            seen_exception_name = self._db_manager.select_seen_exception_for_test_name(
                test_name
            )
            passing_predicate_instance_sequence = self._run_predicate_sequence(
                temp_project_path,
                test_name,
                indexed_pred_seq_str,
                target_src,
                exclude_list,
                seen_exception_name,
                timeout_limit,
            )

            self._get_test_scored_entity_store_db(
                test_name, passing_predicate_instance_sequence, fl_granularity
            )

        self._remove_temp_project(temp_project_path)

        all_tests_scored_entities = self._get_all_tests_top_n_scored_entities(top_n)

        return all_tests_scored_entities



import json

from fauxpy.fault_localization.collect_mbfl.db_manager import CollectMbflDBManager


class CollectMbflReporter:
    def __init__(self, db_manager: CollectMbflDBManager):
        self._db_manager = db_manager

    # def save_test_cases(self):
    #     test_cases_table = self._db_manager.select_all_test_cases()
    #     json_test_cases_table = json.dumps(test_cases_table)
    #     common.saveInCollectModeTestCaseTable(json_test_cases_table)

    def get_json_test_case_table(self):
        test_cases_table = self._db_manager.select_all_test_cases()
        json_test_cases_table = json.dumps(test_cases_table)
        return json_test_cases_table



import ast
from typing import List, Tuple, Optional

import astor

from fauxpy.fault_localization.ps.ast_lib.instrumentation_transformer import (
    InstrumentationTransformer,
)


class InstrumentationManager:
    @staticmethod
    def _add_instrumentation_import(ast_tree):
        def is_from_future(x) -> bool:
            if isinstance(x, ast.ImportFrom):
                return x.module == "__future__"
            return False

        # Probably it is OK for finding docstrings
        # at the beginning of a module.
        def is_doc_string(x) -> bool:
            if isinstance(x, ast.Expr) and (
                (isinstance(x.value, ast.Str) and isinstance(x.value.s, str))
                or (  # for Python 3.6 and Python 3.7
                    isinstance(x.value, ast.Constant) and isinstance(x.value.value, str)
                )
            ):  # for Python 3.8 and Python 3.9
                return True
            return False

        index = 0
        for index, item in enumerate(ast_tree.body):
            if not is_from_future(item) and not is_doc_string(item):
                break

        new_import = ast.ImportFrom(
            module="fauxpy", names=[ast.alias(name="fauxpy_inst", asname=None)], level=0
        )
        ast_tree.body.insert(index, new_import)

    def instrument_current_file_path(
        self,
        file_path: str,
        candidate_predicates: List[Tuple[int, int, str]],
        seen_exceptions: List[Tuple[int, str]],
    ) -> Optional[str]:
        with open(file_path, "r") as source:
            tree = ast.parse(source.read())

        ast_transformer = InstrumentationTransformer(
            tree, candidate_predicates, seen_exceptions
        )
        new_ast = ast_transformer.visit(tree)
        self._add_instrumentation_import(new_ast)
        ast.fix_missing_locations(new_ast)
        try:
            new_ast_content_as_text = astor.to_source(new_ast)
        except AssertionError:
            return None
        return new_ast_content_as_text



# TODO: Consider adding assignment predicates such as x = 3 if p1 else 4.
import ast
from _ast import Expression, While
from typing import Any


class Analyzer(ast.NodeVisitor):
    def __init__(self, line_number: int):
        self._lineNumber = line_number
        self._lineStart = -1
        self._lineEnd = -1

    def get_line_start(self):
        return self._lineStart

    def get_line_end(self):
        return self._lineEnd

    def visit_If(self, node: Expression) -> Any:
        if node.test.lineno <= self._lineNumber:
            line_end = self._get_ending_line(node.test)
            if self._lineNumber <= line_end:
                self._lineStart = node.test.lineno
                self._lineEnd = line_end

        self.generic_visit(node)

    def visit_While(self, node: While) -> Any:
        if node.test.lineno <= self._lineNumber:
            line_end = self._get_ending_line(node.test)
            if self._lineNumber <= line_end:
                self._lineStart = node.test.lineno
                self._lineEnd = line_end

        self.generic_visit(node)

    # TODO: if node.end_lineno exists, use it. Otherwise, the following one.
    def _get_ending_line(self, node):
        line_end_max = -1

        if hasattr(node, "__dict__"):
            all_att = node.__dict__
            for att in all_att:
                att_instance = node.__getattribute__(att)
                if isinstance(att_instance, list):
                    for subAtt in att_instance:
                        line_end_max = max(self._get_ending_line(subAtt), line_end_max)
                else:
                    line_end_max = max(
                        self._get_ending_line(att_instance), line_end_max
                    )
                if hasattr(att_instance, "lineno"):
                    line_end_max = max(att_instance.lineno, line_end_max)

        if hasattr(node, "lineno"):
            line_end_max = max(node.lineno, line_end_max)
        return line_end_max



import ast
import logging
from typing import Tuple, Optional, List

from fauxpy.fault_localization.granularity.ast_lib import FunctionVisitor
from fauxpy.fault_localization.granularity.db_manager import FunctionLevelDbManager


class FunctionLevelGranularity:
    def __init__(self, db_manager: FunctionLevelDbManager):
        self._db_manager = db_manager

    @staticmethod
    def _find_all_function_ranges(file_path: str) -> List[Tuple[str, int, int]]:
        with open(file_path, "r") as source:
            try:
                tree = ast.parse(source.read())
            except Exception as exp:
                logging.warning(
                    f"Due to the following error, the following module is removed form fault localization."
                    f"To include it, fix the error, and run the tool again.\r\n"
                    f"Error: {exp}.\r\n"
                    f"Module: {file_path}"
                )
                return []

        analyzer = FunctionVisitor()
        analyzer.visit(tree)

        return analyzer.function_range_list

    def get_covered_function(
        self, file_path: str, line_number: int
    ) -> Optional[Tuple[str, str, int, int]]:
        function_range_list = self._db_manager.select_function_ranges(
            file_path, line_number
        )

        if len(function_range_list) == 0:
            if self._db_manager.is_file_path_covered(file_path):
                return None
            else:
                function_range_list = self._find_all_function_ranges(file_path)
                for functionRange in function_range_list:
                    self._db_manager.insert_function_information(
                        file_path, functionRange[0], functionRange[1], functionRange[2]
                    )

                if len(function_range_list) == 0:
                    return None
                else:
                    function_range_list = self._db_manager.select_function_ranges(
                        file_path, line_number
                    )
                    if len(function_range_list) == 0:
                        return None

                    function_range_list.sort(key=lambda x: x[2] - x[1])
                    return (
                        file_path,
                        function_range_list[0][0],
                        function_range_list[0][1],
                        function_range_list[0][2],
                    )
        else:
            function_range_list.sort(key=lambda x: x[2] - x[1])
            return (
                file_path,
                function_range_list[0][0],
                function_range_list[0][1],
                function_range_list[0][2],
            )



from abc import abstractmethod
from typing import List

from .fl_type import FlGranularity, FlFamily
from .fl_session import FlSession
from .fauxpy_path import FauxpyPath
from .targeted_failing_tst import TargetedFailingTst


class FlFamilySession(FlSession):
    @abstractmethod
    def get_fl_granularity(self) -> FlGranularity:
        """
        Returns the fault localization granularity.

        Returns:
            FlGranularity: The fault localization granularity option.
        """
        pass

    @abstractmethod
    def get_fl_family(self) -> FlFamily:
        """
        Returns the fault localization family.

        Returns:
            FlFamily: The fault localization family option.
        """
        pass

    @abstractmethod
    def get_project_working_directory(self) -> FauxpyPath:
        """
        Returns the working directory for the project.

        Returns:
            FauxpyPath: The project working directory.
        """
        pass

    @abstractmethod
    def get_targeted_failing_test_list(self) -> List[TargetedFailingTst]:
        """
        Returns the list of targeted failing tests for fault localization.

        Returns:
            list: A list of targeted failing tests.
        """
        pass



from pathlib import Path
from typing import List

import coverage

from fauxpy import constants
from fauxpy.command_line.pytest_mode import legacy_input
from fauxpy.fault_localization.granularity.db_manager import FunctionLevelDbManager
from fauxpy.fault_localization.granularity.function_level import (
    FunctionLevelGranularity,
)
from fauxpy.fault_localization.sbfl.covered_function_manager import (
    CoveredFunctionManager,
)
from fauxpy.fault_localization.sbfl.db_manager import SbflDbManager
from fauxpy.fault_localization.sbfl.ranking_metric_manager import RankingMetricManager
from fauxpy.fault_localization.util.path_util import PathUtil
from fauxpy.session_lib import naming_lib
from fauxpy.session_lib.fl_family_session import FlFamilySession
from fauxpy.session_lib.fl_type import FlGranularity, FlFamily
from fauxpy.session_lib.fauxpy_path import FauxpyPath
from fauxpy.session_lib.pytest_tst_item import PytestTstItem
from fauxpy.session_lib.targeted_failing_tst import TargetedFailingTst


class SbflSession(FlFamilySession):
    def __init__(
        self,
        target_src: FauxpyPath,
        exclude_list: List[FauxpyPath],
        fl_granularity: FlGranularity,
        top_n: int,
        targeted_failing_test_list: List[TargetedFailingTst],
        report_directory_path: Path,
        project_working_directory: Path,
    ):

        self._target_src = target_src
        self._exclude_list = exclude_list
        self._fl_granularity = fl_granularity
        self._top_n = top_n
        self._targeted_failing_test_list = targeted_failing_test_list

        # TODO: Good idea to change the data base file name
        #  like this _Cov = coverage.Coverage(data_suffix="fauxpy").
        #  Apply it to MBFL and PS as well.
        #  Test it for Python 3.6, 3.7, and 3.8.
        #  The API of Coverage.py changes in different version.
        self._coverage_manager = coverage.Coverage()

        self._current_test_name = None
        self._db_manager = SbflDbManager(report_directory_path)
        self._ranking_metric_manager = RankingMetricManager(self._db_manager)
        self._function_level_db_manager = FunctionLevelDbManager(report_directory_path)
        self._function_level_granularity_manager = FunctionLevelGranularity(
            self._function_level_db_manager
        )
        self._covered_function_manager = CoveredFunctionManager(
            self._function_level_granularity_manager
        )
        self._project_working_directory = project_working_directory
        self._path_util = PathUtil(project_working_directory)

    def __str__(self):
        return "SBFL session"

    def get_fl_granularity(self) -> FlGranularity:
        return self._fl_granularity

    def get_fl_family(self) -> FlFamily:
        return FlFamily.Sbfl

    def get_project_working_directory(self) -> FauxpyPath:
        return FauxpyPath.from_relative_path(str(self._project_working_directory), ".")

    def get_targeted_failing_test_list(self) -> List[TargetedFailingTst]:
        """
        Returns the list of targeted failing tests for fault localization.

        Returns:
            list: A list of targeted failing tests.
        """
        return self._targeted_failing_test_list

    def run_test_call(self, item):
        self._current_test_name = PytestTstItem(item).get_test_name()
        self._coverage_manager.start()

    def run_test_make_report(self, item, call):
        if call.when == constants.PYTEST_CALL:
            test_name = PytestTstItem(item).get_test_name()
            if test_name != self._current_test_name:
                raise Exception(
                    f"Expected to stop coverage for '{self._current_test_name}', but found '{test_name}' instead."
                )
            self._coverage_manager.stop()
            coverage_data = self._coverage_manager.get_data()
            files_covered = coverage_data.measured_files()

            covered_statement_list = []
            for file in files_covered:
                if self._path_util.path_should_be_localized(
                    self._target_src,
                    self._exclude_list,
                    file,
                ):
                    line_list = coverage_data.lines(file)
                    for line in line_list:
                        covered_statement_list.append((file, line))
            if len(covered_statement_list) == 0:
                self._db_manager.insert_empty_test(test_name)
            else:
                if self._fl_granularity == FlGranularity.Statement:
                    covered_statement_name_list = [
                        naming_lib.get_statement_name(x[0], x[1])
                        for x in covered_statement_list
                    ]
                    self._db_manager.insert_execution_trace(
                        test_name, covered_statement_name_list
                    )
                elif self._fl_granularity == FlGranularity.Function:
                    covered_function_name_list = (
                        self._covered_function_manager.get_covered_function_names(
                            covered_statement_list
                        )
                    )
                    self._db_manager.insert_execution_trace(
                        test_name, covered_function_name_list
                    )
                else:
                    # TODO: Probably can be removed. We have input validation now.
                    raise Exception(
                        f"Granularity '{self._fl_granularity.name.lower()}' is not supported."
                    )

            self._coverage_manager.erase()

    def terminal_summary(self, terminal_reporter, exit_status):
        for key, value in terminal_reporter.stats.items():
            if key in [constants.PYTEST_PASSED, constants.PYTEST_FAILED]:
                for test_report in value:
                    test_information = PytestTstItem(test_report)
                    test_path = test_information.get_path()
                    test_method_name = test_information.get_method_name()

                    target = False
                    if (
                        legacy_input.get_targeted_failing_test_list_legacy(
                            self._targeted_failing_test_list
                        )
                        is not None
                        and key == constants.PYTEST_FAILED
                    ):
                        target = legacy_input.get_targeted_failing_test_list_legacy(
                            self._targeted_failing_test_list
                        ).is_target_test(test_path, test_method_name)
                    elif (
                        legacy_input.get_targeted_failing_test_list_legacy(
                            self._targeted_failing_test_list
                        )
                        is None
                        and key == constants.PYTEST_FAILED
                    ):
                        target = True

                    test_name = test_information.get_test_name()
                    self._db_manager.insert_test_case(test_name, key, target)

        score_entity_list = self._ranking_metric_manager.compute_sorted_scores(
            self._top_n
        )

        self._db_manager.end()
        self._function_level_db_manager.end()

        return score_entity_list
